from django.urls import reverse

from ....acl import ACL_CACHE
from ....acl.models import Role
from ....acl.admin.test import mock_role_form_data
from ....admin.test import AdminTestCase
from ....cache.test import assert_invalidates_cache
from ...models import Category, CategoryRole


def create_data(data_dict):
    return mock_role_form_data(CategoryRole(), data_dict)


class CategoryRoleAdminViewsTests(AdminTestCase):
    def test_link_registered(self):
        """admin nav contains category roles link"""
        response = self.client.get(reverse("misago:admin:permissions:categories:index"))

        self.assertContains(
            response, reverse("misago:admin:permissions:categories:index")
        )

    def test_list_view(self):
        """roles list view returns 200"""
        response = self.client.get(reverse("misago:admin:permissions:categories:index"))

        self.assertEqual(response.status_code, 200)

    def test_new_view(self):
        """new role view has no showstoppers"""
        response = self.client.get(reverse("misago:admin:permissions:categories:new"))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test CategoryRole"}),
        )
        self.assertEqual(response.status_code, 302)

        test_role = CategoryRole.objects.get(name="Test CategoryRole")
        response = self.client.get(reverse("misago:admin:permissions:categories:index"))
        self.assertContains(response, test_role.name)

    def test_edit_view(self):
        """edit role view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test CategoryRole"}),
        )

        test_role = CategoryRole.objects.get(name="Test CategoryRole")

        response = self.client.get(
            reverse(
                "misago:admin:permissions:categories:edit", kwargs={"pk": test_role.pk}
            )
        )
        self.assertContains(response, "Test CategoryRole")

        response = self.client.post(
            reverse(
                "misago:admin:permissions:categories:edit", kwargs={"pk": test_role.pk}
            ),
            data=create_data({"name": "Top Lel"}),
        )
        self.assertEqual(response.status_code, 302)

        test_role = CategoryRole.objects.get(name="Top Lel")
        response = self.client.get(reverse("misago:admin:permissions:categories:index"))
        self.assertContains(response, test_role.name)

    def test_editing_role_invalidates_acl_cache(self):
        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test CategoryRole"}),
        )

        test_role = CategoryRole.objects.get(name="Test CategoryRole")

        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse(
                    "misago:admin:permissions:categories:edit",
                    kwargs={"pk": test_role.pk},
                ),
                data=create_data({"name": "Top Lel"}),
            )

    def test_delete_view(self):
        """delete role view has no showstoppers"""
        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test CategoryRole"}),
        )

        test_role = CategoryRole.objects.get(name="Test CategoryRole")
        response = self.client.post(
            reverse(
                "misago:admin:permissions:categories:delete",
                kwargs={"pk": test_role.pk},
            )
        )
        self.assertEqual(response.status_code, 302)

        self.client.get(reverse("misago:admin:permissions:categories:index"))
        response = self.client.get(reverse("misago:admin:permissions:categories:index"))
        self.assertNotContains(response, test_role.name)

    def test_deleting_role_invalidates_acl_cache(self):
        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test CategoryRole"}),
        )

        test_role = CategoryRole.objects.get(name="Test CategoryRole")

        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse(
                    "misago:admin:permissions:categories:delete",
                    kwargs={"pk": test_role.pk},
                )
            )

    def test_change_category_roles_view(self):
        """change category roles perms view works"""
        root = Category.objects.root_category()
        for descendant in root.get_descendants():
            descendant.delete()

        # Create categories tree for test cases:
        #
        # Category A
        #   + Category B
        # Category C
        #   + Category D

        root = Category.objects.root_category()
        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category A",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        test_category = Category.objects.get(slug="category-a")

        self.assertEqual(Category.objects.count(), 3)

        # Create test roles
        self.client.post(
            reverse("misago:admin:permissions:new"),
            data=mock_role_form_data(Role(), {"name": "Test Role A"}),
        )
        self.client.post(
            reverse("misago:admin:permissions:new"),
            data=mock_role_form_data(Role(), {"name": "Test Role B"}),
        )

        test_role_a = Role.objects.get(name="Test Role A")
        test_role_b = Role.objects.get(name="Test Role B")

        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test Comments"}),
        )
        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test Full"}),
        )

        role_comments = CategoryRole.objects.get(name="Test Comments")
        role_full = CategoryRole.objects.get(name="Test Full")

        # Test view itself
        # See if form page is rendered
        response = self.client.get(
            reverse(
                "misago:admin:categories:permissions", kwargs={"pk": test_category.pk}
            )
        )
        self.assertContains(response, test_category.name)
        self.assertContains(response, test_role_a.name)
        self.assertContains(response, test_role_b.name)
        self.assertContains(response, role_comments.name)
        self.assertContains(response, role_full.name)

        # Assign roles to categories
        response = self.client.post(
            reverse(
                "misago:admin:categories:permissions", kwargs={"pk": test_category.pk}
            ),
            data={
                ("%s-category_role" % test_role_a.pk): role_full.pk,
                ("%s-category_role" % test_role_b.pk): role_comments.pk,
            },
        )
        self.assertEqual(response.status_code, 302)

        # Check that roles were assigned
        category_role_set = test_category.category_role_set
        self.assertEqual(
            category_role_set.get(role=test_role_a).category_role_id, role_full.pk
        )
        self.assertEqual(
            category_role_set.get(role=test_role_b).category_role_id, role_comments.pk
        )

        # Check that ACL was invalidated
        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse(
                    "misago:admin:categories:permissions",
                    kwargs={"pk": test_category.pk},
                ),
                data={
                    ("%s-category_role" % test_role_a.pk): role_full.pk,
                    ("%s-category_role" % test_role_b.pk): role_comments.pk,
                },
            )

    def test_change_role_categories_permissions_view(self):
        """change role categories perms view works"""
        self.client.post(
            reverse("misago:admin:permissions:new"),
            data=mock_role_form_data(Role(), {"name": "Test CategoryRole"}),
        )

        test_role = Role.objects.get(name="Test CategoryRole")

        root = Category.objects.root_category()
        for descendant in root.get_descendants():
            descendant.delete()

        self.assertEqual(Category.objects.count(), 2)
        response = self.client.get(
            reverse("misago:admin:permissions:categories", kwargs={"pk": test_role.pk})
        )
        self.assertEqual(response.status_code, 302)

        # Create categories tree for test cases:
        #
        # Category A
        #   + Category B
        # Category C
        #   + Category D

        root = Category.objects.root_category()
        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category A",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category C",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )

        category_a = Category.objects.get(slug="category-a")
        category_c = Category.objects.get(slug="category-c")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category B",
                "new_parent": category_a.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        category_b = Category.objects.get(slug="category-b")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category D",
                "new_parent": category_c.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        category_d = Category.objects.get(slug="category-d")

        self.assertEqual(Category.objects.count(), 6)

        # See if form page is rendered
        response = self.client.get(
            reverse("misago:admin:permissions:categories", kwargs={"pk": test_role.pk})
        )
        self.assertContains(response, category_a.name)
        self.assertContains(response, category_b.name)
        self.assertContains(response, category_c.name)
        self.assertContains(response, category_d.name)

        # Set test roles
        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test Comments"}),
        )
        role_comments = CategoryRole.objects.get(name="Test Comments")

        self.client.post(
            reverse("misago:admin:permissions:categories:new"),
            data=create_data({"name": "Test Full"}),
        )
        role_full = CategoryRole.objects.get(name="Test Full")

        # See if form contains those roles
        response = self.client.get(
            reverse("misago:admin:permissions:categories", kwargs={"pk": test_role.pk})
        )
        self.assertContains(response, role_comments.name)
        self.assertContains(response, role_full.name)

        # Assign roles to categories
        response = self.client.post(
            reverse("misago:admin:permissions:categories", kwargs={"pk": test_role.pk}),
            data={
                ("%s-role" % category_a.pk): role_comments.pk,
                ("%s-role" % category_b.pk): role_comments.pk,
                ("%s-role" % category_c.pk): role_full.pk,
                ("%s-role" % category_d.pk): role_full.pk,
            },
        )
        self.assertEqual(response.status_code, 302)

        # Check that roles were assigned
        categories_acls = test_role.categories_acls
        self.assertEqual(
            categories_acls.get(category=category_a).category_role_id, role_comments.pk
        )
        self.assertEqual(
            categories_acls.get(category=category_b).category_role_id, role_comments.pk
        )
        self.assertEqual(
            categories_acls.get(category=category_c).category_role_id, role_full.pk
        )
        self.assertEqual(
            categories_acls.get(category=category_d).category_role_id, role_full.pk
        )

        # Check that ACL was invalidated
        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse(
                    "misago:admin:permissions:categories", kwargs={"pk": test_role.pk}
                ),
                data={
                    ("%s-role" % category_a.pk): role_comments.pk,
                    ("%s-role" % category_b.pk): role_comments.pk,
                    ("%s-role" % category_c.pk): role_full.pk,
                    ("%s-role" % category_d.pk): role_full.pk,
                },
            )



from django.utils.translation import gettext_lazy as _


class OAuth2Error(Exception):
    pass


class OAuth2ProviderError(OAuth2Error):
    pass


class OAuth2AccessDeniedError(OAuth2ProviderError):
    message = _("The OAuth2 process was canceled by the provider.")


class OAuth2StateError(OAuth2Error):
    recoverable = True


class OAuth2StateNotSetError(OAuth2StateError):
    message = _("The OAuth2 session is missing state.")


class OAuth2StateNotProvidedError(OAuth2StateError):
    message = _("The OAuth2 state was not sent by the provider.")


class OAuth2StateMismatchError(OAuth2StateError):
    message = _(
        "The OAuth2 state sent by the provider did not match one in the session."
    )


class OAuth2CodeError(OAuth2Error):
    recoverable = True


class OAuth2CodeNotProvidedError(OAuth2CodeError):
    message = _("The OAuth2 authorization code was not sent by the provider.")


class OAuth2ProviderError(OAuth2Error):
    recoverable = True


class OAuth2AccessTokenRequestError(OAuth2ProviderError):
    message = _("Failed to connect to the OAuth2 provider to retrieve an access token.")


class OAuth2AccessTokenResponseError(OAuth2ProviderError):
    message = _("The OAuth2 provider responded with error for an access token request.")


class OAuth2AccessTokenJSONError(OAuth2ProviderError):
    message = _(
        "The OAuth2 provider did not respond with a valid JSON "
        "for an access token request."
    )


class OAuth2AccessTokenNotProvidedError(OAuth2ProviderError):
    message = _("JSON sent by the OAuth2 provider did not contain an access token.")


class OAuth2UserDataRequestError(OAuth2ProviderError):
    message = _("Failed to connect to the OAuth2 provider to retrieve user profile.")


class OAuth2UserDataResponseError(OAuth2ProviderError):
    message = _("The OAuth2 provider responded with error for user profile request.")


class OAuth2UserDataJSONError(OAuth2ProviderError):
    message = _(
        "The OAuth2 provider did not respond with a valid JSON "
        "for user profile request."
    )


class OAuth2UserIdNotProvidedError(OAuth2Error):
    message = _("JSON sent by the OAuth2 provider did not contain a user id.")


class OAuth2UserAccountDeactivatedError(OAuth2Error):
    recoverable = False
    message = _(
        "User account associated with the profile from the OAuth2 provider was "
        "deactivated by the site administrator."
    )


class OAuth2UserDataValidationError(OAuth2ProviderError):
    recoverable = False
    error_list: list[str]
    message = _("User profile retrieved from the OAuth2 provider did not validate.")

    def __init__(self, error_list: list[str]):
        self.error_list = error_list



from django.dispatch import Signal, receiver

from ..categories import PRIVATE_THREADS_ROOT_NAME
from ..categories.signals import delete_category_content, move_category_content
from ..threads.signals import merge_post, merge_thread, move_post, move_thread

thread_read = Signal()


@receiver(delete_category_content)
def delete_category_threads(sender, **kwargs):
    sender.postread_set.all().delete()


@receiver(move_category_content)
def move_category_tracker(sender, **kwargs):
    sender.postread_set.update(category=kwargs["new_category"])


@receiver(merge_thread)
def merge_thread_tracker(sender, **kwargs):
    other_thread = kwargs["other_thread"]
    other_thread.postread_set.update(category=sender.category, thread=sender)


@receiver(move_thread)
def move_thread_tracker(sender, **kwargs):
    sender.postread_set.update(category=sender.category, thread=sender)


@receiver(merge_post)
def merge_post_delete_tracker(sender, **kwargs):
    sender.postread_set.all().delete()


@receiver(move_post)
def move_post_delete_tracker(sender, **kwargs):
    sender.postread_set.all().delete()


@receiver(thread_read)
def decrease_unread_private_count(sender, **kwargs):
    user = sender
    thread = kwargs["thread"]

    if thread.category.thread_type.root_name != PRIVATE_THREADS_ROOT_NAME:
        return

    if user.unread_private_threads:
        user.unread_private_threads -= 1
        user.save(update_fields=["unread_private_threads"])



from django.urls import path

from ..views import privacy_policy, terms_of_service

urlpatterns = [
    path("privacy-policy/", privacy_policy, name="privacy-policy"),
    path("terms-of-service/", terms_of_service, name="terms-of-service"),
]



from io import BytesIO

import requests
from PIL import Image

from . import store
from ...conf import settings

GRAVATAR_URL = "http://www.gravatar.com/avatar/%s?s=%s&d=404"


class GravatarError(RuntimeError):
    pass


class NoGravatarAvailable(GravatarError):
    pass


def set_avatar(user):
    url_formats = (user.email_hash, max(settings.MISAGO_AVATARS_SIZES))
    try:
        r = requests.get(GRAVATAR_URL % url_formats, timeout=5)
        if r.status_code != 200:
            raise NoGravatarAvailable("gravatar is not available for this e-mail")

        image = Image.open(BytesIO(r.content))
        store.store_new_avatar(user, image)
    except requests.exceptions.RequestException:
        raise GravatarError("failed to connect to gravatar servers")



import pytest
from django.urls import reverse

from ....admin.test import AdminTestCase
from ....cache.test import assert_invalidates_cache
from ....test import assert_contains
from ... import ACL_CACHE
from ...models import Role
from ..test import mock_role_form_data

admin_link = reverse("misago:admin:permissions:index")


def create_form_data(data_dict):
    return mock_role_form_data(Role(), data_dict)


def test_link_is_registered_in_admin_nav(admin_client):
    response = admin_client.get(reverse("misago:admin:index"))
    assert_contains(response, admin_link)


def test_list_renders(admin_client):
    response = admin_client.get(admin_link)
    assert response.status_code == 200


def test_new_role_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:permissions:new"))
    assert response.status_code == 200


def test_new_role_can_be_created(admin_client):
    response = admin_client.post(
        reverse("misago:admin:permissions:new"),
        data=create_form_data({"name": "Test Role"}),
    )

    Role.objects.get(name="Test Role")


@pytest.fixture
def role(db):
    return Role.objects.create(name="Test Role")


def test_edit_role_form_renders(admin_client, role):
    response = admin_client.get(
        reverse("misago:admin:permissions:edit", kwargs={"pk": role.pk})
    )
    assert response.status_code == 200


def test_role_can_be_edited(admin_client, role):
    response = admin_client.post(
        reverse("misago:admin:permissions:edit", kwargs={"pk": role.pk}),
        data=create_form_data({"name": "Edited Role"}),
    )

    role.refresh_from_db()
    assert role.name == "Edited Role"


def test_editing_role_invalidates_acl_cache(admin_client, role):
    with assert_invalidates_cache(ACL_CACHE):
        admin_client.post(
            reverse("misago:admin:permissions:edit", kwargs={"pk": role.pk}),
            data=create_form_data({"name": "Role"}),
        )


def test_role_can_be_deleted(admin_client, role):
    admin_client.post(
        reverse("misago:admin:permissions:delete", kwargs={"pk": role.pk})
    )

    with pytest.raises(Role.DoesNotExist):
        role.refresh_from_db()


def test_special_role_cant_be_deleted(admin_client, role):
    role.special_role = "Test"
    role.save()

    admin_client.post(
        reverse("misago:admin:permissions:delete", kwargs={"pk": role.pk})
    )

    role.refresh_from_db()


def test_deleting_role_invalidates_acl_cache(admin_client, role):
    with assert_invalidates_cache(ACL_CACHE):
        admin_client.post(
            reverse("misago:admin:permissions:delete", kwargs={"pk": role.pk})
        )


def test_users_with_role_view_redirects_to_admin_users_list(admin_client, role):
    response = admin_client.get(
        reverse("misago:admin:permissions:users", kwargs={"pk": role.pk})
    )
    assert response.status_code == 302
    assert reverse("misago:admin:users:index") in response["location"]



import pytest
from django.urls import reverse


@pytest.fixture
def users_admin_link(admin_client):
    response = admin_client.get(reverse("misago:admin:users:index"))
    return response["location"]



from django.urls import reverse

from ...acl.test import patch_user_acl
from ..test import AuthenticatedUserTestCase, create_test_user


class UserEditDetailsApiTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.api_link = reverse(
            "misago:api:user-edit-details", kwargs={"pk": self.user.pk}
        )

    def get_profile_fields(self):
        self.user.refresh_from_db()
        return self.user.profile_fields

    def test_api_has_no_showstoppers(self):
        """api outputs response for freshly created user"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

    def test_api_has_no_showstoppers_old_user(self):
        """api outputs response for freshly created user"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

    def test_other_user(self):
        """api handles scenario when its other user looking at profile"""
        test_user = create_test_user("OtherUser", "otheruser@example.com")

        api_link = reverse("misago:api:user-edit-details", kwargs={"pk": test_user.pk})

        # moderator has permission to edit details
        with patch_user_acl({"can_moderate_profile_details": True}):
            response = self.client.get(api_link)
            self.assertEqual(response.status_code, 200)

        # non-moderator has no permission to edit details
        with patch_user_acl({"can_moderate_profile_details": False}):
            response = self.client.get(api_link)
            self.assertEqual(response.status_code, 403)

    def test_nonexistant_user(self):
        """api handles nonexistant users"""
        api_link = reverse(
            "misago:api:user-edit-details", kwargs={"pk": self.user.pk + 123}
        )

        response = self.client.get(api_link)
        self.assertEqual(response.status_code, 404)

    def test_api_updates_text_field(self):
        """api updates text field"""
        response = self.client.post(
            self.api_link, data={"bio": "I have some, as is tradition."}
        )
        self.assertEqual(response.status_code, 200)

        profile_fields = self.get_profile_fields()
        self.assertEqual(profile_fields["bio"], "I have some, as is tradition.")

        response_json = response.json()
        self.assertEqual(response_json["id"], self.user.id)
        self.assertTrue(response_json["edit"])
        self.assertTrue(response_json["groups"])

    def test_api_updates_select_field(self):
        """api updates select field"""
        response = self.client.post(self.api_link, data={"gender": "female"})

        self.assertEqual(response.status_code, 200)

        profile_fields = self.get_profile_fields()
        self.assertEqual(profile_fields["gender"], "female")

        response_json = response.json()
        self.assertEqual(response_json["id"], self.user.id)
        self.assertTrue(response_json["edit"])
        self.assertTrue(response_json["groups"])

    def test_api_validates_url_field(self):
        """api runs basic validation against url fields"""
        response = self.client.post(self.api_link, data={"website": "noturl"})

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"website": ["Enter a valid URL."]})

    def test_api_cleans_url_field(self):
        """api cleans url fields"""
        response = self.client.post(self.api_link, data={"website": "onet.pl"})

        self.assertEqual(response.status_code, 200)

        profile_fields = self.get_profile_fields()
        self.assertEqual(profile_fields["website"], "http://onet.pl")

        response_json = response.json()
        self.assertEqual(response_json["id"], self.user.id)
        self.assertTrue(response_json["edit"])
        self.assertTrue(response_json["groups"])

    def test_api_custom_cleans_url_field(self):
        """api calls fields clean method"""
        response = self.client.post(self.api_link, data={"twitter": "@Weebl"})

        self.assertEqual(response.status_code, 200)

        profile_fields = self.get_profile_fields()
        self.assertEqual(profile_fields["twitter"], "Weebl")

        response_json = response.json()
        self.assertEqual(response_json["id"], self.user.id)
        self.assertTrue(response_json["edit"])
        self.assertTrue(response_json["groups"])



from django.test import TestCase

from ...users import credentialchange
from ..test import create_test_user


class MockRequest:
    def __init__(self, user):
        self.session = {}
        self.user = user


class CredentialChangeTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")

    def test_valid_token_generation(self):
        """credentialchange module allows for store and read of change token"""
        request = MockRequest(self.user)
        token = credentialchange.store_new_credential(
            request, "email", "newmail@example.com"
        )

        email = credentialchange.read_new_credential(request, "email", token)
        self.assertEqual(email, "newmail@example.com")

    def test_email_change_invalidated_token(self):
        """token is invalidated by email change"""
        request = MockRequest(self.user)
        token = credentialchange.store_new_credential(
            request, "email", "newmail@example.com"
        )

        self.user.set_email("otheremail@example.com")
        self.user.save()

        email = credentialchange.read_new_credential(request, "email", token)
        self.assertIsNone(email)

    def test_password_change_invalidated_token(self):
        """token is invalidated by password change"""
        request = MockRequest(self.user)
        token = credentialchange.store_new_credential(
            request, "email", "newmail@example.com"
        )

        self.user.set_password("Egebeg!123")
        self.user.save()

        email = credentialchange.read_new_credential(request, "email", token)
        self.assertIsNone(email)

    def test_invalid_token_is_handled(self):
        """there are no explosions in invalid tokens handling"""
        request = MockRequest(self.user)
        token = credentialchange.store_new_credential(
            request, "email", "newmail@example.com"
        )

        email = credentialchange.read_new_credential(request, "em4il", token)
        self.assertIsNone(email)



from urllib.parse import urlparse

from django.urls import reverse

from ...conf.test import override_dynamic_settings
from ...test import assert_contains
from ...users.bans import ban_ip


def test_oauth2_login_view_returns_404_if_oauth_is_disabled(client, dynamic_settings):
    assert dynamic_settings.enable_oauth2_client is False

    response = client.get(reverse("misago:oauth2-login"))
    assert response.status_code == 404


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_login_view_returns_redirect_302_if_oauth_is_enabled(
    client, dynamic_settings
):
    assert dynamic_settings.enable_oauth2_client is True

    response = client.get(reverse("misago:oauth2-login"))
    assert response.status_code == 302

    redirect_to = urlparse(response["Location"])
    assert redirect_to.netloc == "example.com"
    assert redirect_to.path == "/oauth2/login"
    assert "clientid123" in redirect_to.query


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_login_view_returns_error_403_if_user_ip_is_banned(
    client, dynamic_settings
):
    ban_ip("127.*", "Ya got banned!")

    assert dynamic_settings.enable_oauth2_client is True

    response = client.get(reverse("misago:oauth2-login"))
    assert_contains(response, "Ya got banned", 403)


def test_oauth2_login_view_returns_error_404_if_user_ip_is_banned_and_oauth_is_disabled(
    client, dynamic_settings
):
    ban_ip("127.*", "Ya got banned!")

    assert dynamic_settings.enable_oauth2_client is False

    response = client.get(reverse("misago:oauth2-login"))
    assert response.status_code == 404



from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from ..management.commands import deleteprofilefield
from ..test import create_test_user


class DeleteProfileFieldTests(TestCase):
    def test_no_fieldname(self):
        """utility has no showstoppers when no fieldname is given"""
        out = StringIO()
        call_command(deleteprofilefield.Command(), stderr=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Specify fieldname to delete.")

    def test_no_fields_set(self):
        """utility has no showstoppers when no fields are set"""
        out = StringIO()
        call_command(deleteprofilefield.Command(), "gender", stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(
            command_output, '"gender" profile field has been deleted from 0 users.'
        )

    def test_delete_fields(self):
        """utility has no showstoppers when no fields are set"""
        user = create_test_user(
            "User", "user@example.com", profile_fields={"gender": "male", "bio": "Yup!"}
        )

        out = StringIO()
        call_command(deleteprofilefield.Command(), "gender", stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(
            command_output, '"gender" profile field has been deleted from 1 users.'
        )

        user.refresh_from_db()
        self.assertEqual(user.profile_fields, {"bio": "Yup!"})



from unittest.mock import Mock

from ..online.utils import get_user_status
from ..test import AuthenticatedUserTestCase, create_test_user


class GetUserStatusTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.other_user = create_test_user("OtherUser", "otheruser@example.com")

    def test_get_visible_user_status_returns_online(self):
        request = Mock(
            user=self.user,
            user_acl={"can_see_hidden_users": False},
            cache_versions={"bans": "abcdefgh"},
        )
        assert get_user_status(request, self.other_user)["is_online"]

    def test_get_hidden_user_status_without_seeing_hidden_permission_returns_offline(
        self,
    ):
        """get_user_status has no showstopper for hidden user"""
        self.other_user.is_hiding_presence = True
        self.other_user.save()

        request = Mock(
            user=self.user,
            user_acl={"can_see_hidden_users": False},
            cache_versions={"bans": "abcdefgh"},
        )
        assert get_user_status(request, self.other_user)["is_hidden"]

    def test_get_hidden_user_status_with_seeing_hidden_permission_returns_online_hidden(
        self,
    ):
        self.other_user.is_hiding_presence = True
        self.other_user.save()

        request = Mock(
            user=self.user,
            user_acl={"can_see_hidden_users": True},
            cache_versions={"bans": "abcdefgh"},
        )
        assert get_user_status(request, self.other_user)["is_online_hidden"]



import json


class GraphQLTestClient:
    def __init__(self, client, url):
        self.client = client
        self.url = url

    def query(self, query, variables=None):
        data = {"query": query}
        if variables:
            data["variables"] = variables
        response = self.client.post(
            self.url, json.dumps(data), content_type="application/json"
        )
        json_data = response.json()
        assert not json_data.get("errors"), json_data.get("errors")
        return json_data["data"]



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

from ...acl.models import Role
from ...admin.forms import IsoDateTimeField, YesNoSwitch
from ...conf.admin.forms import ChangeSettingsForm
from ...core.validators import validate_sluggable
from ...search.filter_queryset import filter_queryset
from ..models import Ban, DataDownload, Rank
from ..profilefields import profilefields
from ..utils import hash_email
from ..validators import validate_email, validate_username

User = get_user_model()


class UserBaseForm(forms.ModelForm):
    username = forms.CharField(label=_("Username"))
    title = forms.CharField(label=_("Custom title"), required=False)
    email = forms.EmailField(label=_("E-mail address"))

    class Meta:
        model = User
        fields = ["username", "email", "title"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        self.settings = self.request.settings

        super().__init__(*args, **kwargs)

    def clean_username(self):
        data = self.cleaned_data["username"]
        validate_username(self.settings, data, exclude=self.instance)
        return data

    def clean_email(self):
        data = self.cleaned_data["email"]
        validate_email(data, exclude=self.instance)
        return data

    def clean_new_password(self):
        data = self.cleaned_data["new_password"]
        if data:
            validate_password(data, user=self.instance)
        return data

    def clean_roles(self):
        data = self.cleaned_data["roles"]

        for role in data:
            if role.special_role == "authenticated":
                break
        else:
            message = _('All registered members must have a "Member" role.')
            raise forms.ValidationError(message)

        return data


class NewUserForm(UserBaseForm):
    new_password = forms.CharField(
        label=_("Password"), strip=False, widget=forms.PasswordInput
    )

    class Meta:
        model = User
        fields = ["username", "email", "title"]


class EditUserForm(UserBaseForm):
    IS_STAFF_LABEL = _("Is administrator")
    IS_STAFF_HELP_TEXT = _(
        "Designates whether the user can log into admin sites. "
        "If Django admin site is enabled, this user will need "
        "additional permissions assigned within it to admin "
        "Django modules."
    )

    IS_SUPERUSER_LABEL = _("Is superuser")
    IS_SUPERUSER_HELP_TEXT = _(
        "Only administrators can access admin sites. "
        "In addition to admin site access, superadmins "
        "can also change other members admin levels."
    )

    IS_ACTIVE_LABEL = _("Is active")
    IS_ACTIVE_HELP_TEXT = _(
        "Designates whether this user should be treated as active. "
        "Turning this off is non-destructible way to remove user accounts."
    )

    IS_ACTIVE_STAFF_MESSAGE_LABEL = _("Staff message")
    IS_ACTIVE_STAFF_MESSAGE_HELP_TEXT = _(
        "Optional message for forum team members explaining "
        "why user's account has been disabled."
    )

    new_password = forms.CharField(
        label=_("Change password to"),
        strip=False,
        widget=forms.PasswordInput,
        required=False,
    )

    is_avatar_locked = YesNoSwitch(
        label=_("Lock avatar"),
        help_text=_(
            "Setting this to yes will stop user from changing "
            "his/her avatar, and will reset his/her avatar to "
            "procedurally generated one."
        ),
    )
    avatar_lock_user_message = forms.CharField(
        label=_("User message"),
        help_text=_(
            "Optional message for user explaining "
            "why he/she is banned form changing avatar."
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )
    avatar_lock_staff_message = forms.CharField(
        label=_("Staff message"),
        help_text=_(
            "Optional message for forum team members explaining "
            "why user is banned form changing avatar."
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )

    signature = forms.CharField(
        label=_("Signature contents"),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )
    is_signature_locked = YesNoSwitch(
        label=_("Lock signature"),
        help_text=_(
            "Setting this to yes will stop user from "
            "making changes to his/her signature."
        ),
    )
    signature_lock_user_message = forms.CharField(
        label=_("User message"),
        help_text=_(
            "Optional message to user explaining why his/hers signature is locked."
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )
    signature_lock_staff_message = forms.CharField(
        label=_("Staff message"),
        help_text=_(
            "Optional message to team members explaining why user signature is locked."
        ),
        widget=forms.Textarea(attrs={"rows": 3}),
        required=False,
    )

    is_hiding_presence = YesNoSwitch(label=_("Hides presence"))

    limits_private_thread_invites_to = forms.TypedChoiceField(
        label=_("Who can add user to private threads"),
        coerce=int,
        choices=User.LIMIT_INVITES_TO_CHOICES,
    )

    subscribe_to_started_threads = forms.TypedChoiceField(
        label=_("Started threads"), coerce=int, choices=User.SUBSCRIPTION_CHOICES
    )
    subscribe_to_replied_threads = forms.TypedChoiceField(
        label=_("Replid threads"), coerce=int, choices=User.SUBSCRIPTION_CHOICES
    )

    class Meta:
        model = User
        fields = [
            "username",
            "email",
            "title",
            "is_avatar_locked",
            "avatar_lock_user_message",
            "avatar_lock_staff_message",
            "signature",
            "is_signature_locked",
            "is_hiding_presence",
            "limits_private_thread_invites_to",
            "signature_lock_user_message",
            "signature_lock_staff_message",
            "subscribe_to_started_threads",
            "subscribe_to_replied_threads",
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        profilefields.add_fields_to_admin_form(self.request, self.instance, self)

    def get_profile_fields_groups(self):
        profile_fields_groups = []
        for group in self._profile_fields_groups:
            fields_group = {"name": group["name"], "fields": []}

            for fieldname in group["fields"]:
                fields_group["fields"].append(self[fieldname])

            profile_fields_groups.append(fields_group)
        return profile_fields_groups

    def clean_signature(self):
        data = self.cleaned_data["signature"]

        length_limit = self.settings.signature_length_max
        if len(data) > length_limit:
            message = ngettext(
                "Signature can't be longer than %(limit)s character.",
                "Signature can't be longer than %(limit)s characters.",
                length_limit,
            )
            raise forms.ValidationError(message % {"limit": length_limit})

        return data

    def clean(self):
        data = super().clean()
        return profilefields.clean_form(self.request, self.instance, self, data)


def UserFormFactory(FormType, instance):
    extra_fields = {}

    extra_fields["rank"] = forms.ModelChoiceField(
        label=_("Rank"),
        help_text=_(
            "Ranks are used to group and distinguish users. They are "
            "also used to add permissions to groups of users."
        ),
        queryset=Rank.objects.order_by("name"),
        initial=instance.rank,
    )

    roles = Role.objects.order_by("name")

    extra_fields["roles"] = forms.ModelMultipleChoiceField(
        label=_("Roles"),
        help_text=_(
            'Individual roles of this user. All users must have a "Member" role.'
        ),
        queryset=roles,
        initial=instance.roles.all() if instance.pk else None,
        widget=forms.CheckboxSelectMultiple,
    )

    return type("UserFormFinal", (FormType,), extra_fields)


def StaffFlagUserFormFactory(FormType, instance):
    staff_fields = {
        "is_staff": YesNoSwitch(
            label=EditUserForm.IS_STAFF_LABEL,
            help_text=EditUserForm.IS_STAFF_HELP_TEXT,
            initial=instance.is_staff,
        ),
        "is_superuser": YesNoSwitch(
            label=EditUserForm.IS_SUPERUSER_LABEL,
            help_text=EditUserForm.IS_SUPERUSER_HELP_TEXT,
            initial=instance.is_superuser,
        ),
    }

    return type("StaffUserForm", (FormType,), staff_fields)


def UserIsActiveFormFactory(FormType, instance):
    is_active_fields = {
        "is_active": YesNoSwitch(
            label=EditUserForm.IS_ACTIVE_LABEL,
            help_text=EditUserForm.IS_ACTIVE_HELP_TEXT,
            initial=instance.is_active,
        ),
        "is_active_staff_message": forms.CharField(
            label=EditUserForm.IS_ACTIVE_STAFF_MESSAGE_LABEL,
            help_text=EditUserForm.IS_ACTIVE_STAFF_MESSAGE_HELP_TEXT,
            initial=instance.is_active_staff_message,
            widget=forms.Textarea(attrs={"rows": 3}),
            required=False,
        ),
    }

    return type("UserIsActiveForm", (FormType,), is_active_fields)


def EditUserFormFactory(
    FormType, instance, add_is_active_fields=False, add_admin_fields=False
):
    FormType = UserFormFactory(FormType, instance)

    if add_is_active_fields:
        FormType = UserIsActiveFormFactory(FormType, instance)

    if add_admin_fields:
        FormType = StaffFlagUserFormFactory(FormType, instance)

    return FormType


class BaseFilterUsersForm(forms.Form):
    username = forms.CharField(label=_("Username"), required=False)
    email = forms.CharField(label=_("E-mail"), required=False)
    profilefields = forms.CharField(label=_("Profile fields contain"), required=False)
    is_inactive = forms.BooleanField(label=_("Requires activation"))
    is_disabled = forms.BooleanField(label=_("Account disabled"))
    is_staff = forms.BooleanField(label=_("Administrator"))
    is_deleting_account = forms.BooleanField(label=_("Deletes their account"))

    def filter_queryset(self, criteria, queryset):
        if criteria.get("username"):
            queryset = filter_queryset(
                queryset, "slug", criteria.get("username").lower()
            )

        if criteria.get("email"):
            queryset = filter_queryset(
                queryset, "email", criteria.get("email"), case_sensitive=False
            )

        if criteria.get("rank"):
            queryset = queryset.filter(rank_id=criteria.get("rank"))

        if criteria.get("role"):
            queryset = queryset.filter(roles__id=criteria.get("role"))

        if criteria.get("is_inactive"):
            queryset = queryset.filter(requires_activation__gt=0)

        if criteria.get("is_disabled"):
            queryset = queryset.filter(is_active=False)

        if criteria.get("is_staff"):
            queryset = queryset.filter(is_staff=True)

        if criteria.get("is_deleting_account"):
            queryset = queryset.filter(is_deleting_account=True)

        if criteria.get("profilefields", "").strip():
            queryset = profilefields.search_users(
                criteria.get("profilefields").strip(), queryset
            )

        return queryset


def create_filter_users_form():
    """
    Factory that uses cache for ranks and roles,
    and makes those ranks and roles typed choice fields that play nice
    with passing values via GET
    """
    ranks_choices = [("", _("All ranks"))]
    for rank in Rank.objects.order_by("name").iterator():
        ranks_choices.append((rank.pk, rank.name))

    roles_choices = [("", _("All roles"))]
    for role in Role.objects.order_by("name").iterator():
        roles_choices.append((role.pk, role.name))

    extra_fields = {
        "rank": forms.TypedChoiceField(
            label=_("Has rank"), coerce=int, required=False, choices=ranks_choices
        ),
        "role": forms.TypedChoiceField(
            label=_("Has role"), coerce=int, required=False, choices=roles_choices
        ),
    }

    return type("FilterUsersForm", (BaseFilterUsersForm,), extra_fields)


class RankForm(forms.ModelForm):
    name = forms.CharField(
        label=_("Name"),
        validators=[validate_sluggable()],
        help_text=_(
            "Short and descriptive name of all users with this rank. "
            '"The Team" or "Game Masters" are good examples.'
        ),
    )
    title = forms.CharField(
        label=_("User title"),
        required=False,
        help_text=_(
            "Optional, singular version of rank name displayed by user names. "
            'For example "GM" or "Dev".'
        ),
    )
    description = forms.CharField(
        label=_("Description"),
        max_length=2048,
        required=False,
        widget=forms.Textarea(attrs={"rows": 3}),
        help_text=_(
            "Optional description explaining function or status of "
            "members distincted with this rank."
        ),
    )
    roles = forms.ModelMultipleChoiceField(
        label=_("User roles"),
        widget=forms.CheckboxSelectMultiple,
        queryset=Role.objects.order_by("name"),
        required=False,
        help_text=_("Rank can give additional roles to users with it."),
    )
    css_class = forms.CharField(
        label=_("CSS class"),
        required=False,
        help_text=_(
            "Optional css class added to content belonging to this rank owner."
        ),
    )
    is_tab = YesNoSwitch(
        label=_("Give rank dedicated tab on users list"),
        required=False,
        help_text=_(
            "Selecting this option will make users with this rank easily discoverable "
            "by others through dedicated page on forum users list."
        ),
    )

    class Meta:
        model = Rank
        fields = ["name", "description", "css_class", "title", "roles", "is_tab"]

    def clean_name(self):
        data = self.cleaned_data["name"]
        self.instance.set_name(data)

        unique_qs = Rank.objects.filter(slug=self.instance.slug)
        if self.instance.pk:
            unique_qs = unique_qs.exclude(pk=self.instance.pk)

        if unique_qs.exists():
            raise forms.ValidationError(_("This name collides with other rank."))

        return data


class BanUsersForm(forms.Form):
    ban_type = forms.MultipleChoiceField(
        label=_("Values to ban"), widget=forms.CheckboxSelectMultiple, choices=[]
    )
    user_message = forms.CharField(
        label=_("User message"),
        required=False,
        max_length=1000,
        help_text=_("Optional message displayed to users instead of default one."),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": _("Message can't be longer than 1000 characters.")
        },
    )
    staff_message = forms.CharField(
        label=_("Team message"),
        required=False,
        max_length=1000,
        help_text=_("Optional ban message for moderators and administrators."),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": _("Message can't be longer than 1000 characters.")
        },
    )
    expires_on = IsoDateTimeField(label=_("Expiration date"), required=False)

    def __init__(self, *args, **kwargs):
        users = kwargs.pop("users")

        super().__init__(*args, **kwargs)

        self.fields["ban_type"].choices = [
            ("usernames", _("Usernames")),
            ("emails", _("E-mails")),
            ("domains", _("E-mail domains")),
        ]

        enable_ip_bans = list(filter(None, [u.joined_from_ip for u in users]))
        if enable_ip_bans:
            self.fields["ban_type"].choices += [
                ("ip", _("IP addresses")),
                ("ip_first", _("First segment of IP addresses")),
                ("ip_two", _("First two segments of IP addresses")),
            ]


class BanForm(forms.ModelForm):
    check_type = forms.TypedChoiceField(
        label=_("Check type"), coerce=int, choices=Ban.CHOICES
    )
    registration_only = YesNoSwitch(
        label=_("Restrict this ban to registrations"),
        help_text=_(
            "Changing this to yes will make this ban check be only performed on "
            "registration step. This is good if you want to block certain "
            "registrations like ones from recently comprimised e-mail providers, "
            "without harming existing users."
        ),
    )
    banned_value = forms.CharField(
        label=_("Banned value"),
        max_length=250,
        help_text=_(
            "This value is case-insensitive and accepts asterisk (*) "
            "for rought matches. For example, making IP ban for value "
            '"83.*" will ban all IP addresses beginning with "83.".'
        ),
        error_messages={
            "max_length": _("Banned value can't be longer than 250 characters.")
        },
    )
    user_message = forms.CharField(
        label=_("User message"),
        required=False,
        max_length=1000,
        help_text=_("Optional message displayed to user instead of default one."),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": _("Message can't be longer than 1000 characters.")
        },
    )
    staff_message = forms.CharField(
        label=_("Team message"),
        required=False,
        max_length=1000,
        help_text=_("Optional ban message for moderators and administrators."),
        widget=forms.Textarea(attrs={"rows": 3}),
        error_messages={
            "max_length": _("Message can't be longer than 1000 characters.")
        },
    )
    expires_on = IsoDateTimeField(label=_("Expiration date"), required=False)

    class Meta:
        model = Ban
        fields = [
            "check_type",
            "registration_only",
            "banned_value",
            "user_message",
            "staff_message",
            "expires_on",
        ]

    def clean_banned_value(self):
        data = self.cleaned_data["banned_value"]
        while "**" in data:
            data = data.replace("**", "*")

        if data == "*":
            raise forms.ValidationError(_("Banned value is too vague."))

        return data


class FilterBansForm(forms.Form):
    check_type = forms.ChoiceField(
        label=_("Type"),
        required=False,
        choices=[
            ("", _("All bans")),
            ("names", _("Usernames")),
            ("emails", _("E-mails")),
            ("ips", _("IPs")),
        ],
    )
    value = forms.CharField(label=_("Banned value begins with"), required=False)
    registration_only = forms.ChoiceField(
        label=_("Registration only"),
        required=False,
        choices=[("", _("Any")), ("only", _("Yes")), ("exclude", _("No"))],
    )
    state = forms.ChoiceField(
        label=_("State"),
        required=False,
        choices=[("", _("Any")), ("used", _("Active")), ("unused", _("Expired"))],
    )

    def filter_queryset(self, criteria, queryset):
        if criteria.get("check_type") == "names":
            queryset = queryset.filter(check_type=0)

        if criteria.get("check_type") == "emails":
            queryset = queryset.filter(check_type=1)

        if criteria.get("check_type") == "ips":
            queryset = queryset.filter(check_type=2)

        if criteria.get("value"):
            queryset = queryset.filter(
                banned_value__startswith=criteria.get("value").lower()
            )

        if criteria.get("state") == "used":
            queryset = queryset.filter(is_checked=True)

        if criteria.get("state") == "unused":
            queryset = queryset.filter(is_checked=False)

        if criteria.get("registration_only") == "only":
            queryset = queryset.filter(registration_only=True)

        if criteria.get("registration_only") == "exclude":
            queryset = queryset.filter(registration_only=False)

        return queryset


class RequestDataDownloadsForm(forms.Form):
    user_identifiers = forms.CharField(
        label=_("Usernames or emails"),
        help_text=_(
            "Enter every item in new line. Duplicates will be ignored. "
            "This field is case insensitive. Depending on site configuration and "
            "amount of data to archive it may take up to few days for requests to "
            "complete. E-mail will notification will be sent to every user once their "
            "download is ready."
        ),
        widget=forms.Textarea,
    )

    def clean_user_identifiers(self):
        user_identifiers = self.cleaned_data["user_identifiers"].lower().splitlines()
        user_identifiers = list(filter(bool, user_identifiers))
        user_identifiers = list(set(user_identifiers))

        if len(user_identifiers) > 20:
            raise forms.ValidationError(
                _(
                    "You may not enter more than 20 items at a single time "
                    "(You have entered %(show_value)s)."
                )
                % {"show_value": len(user_identifiers)}
            )

        return user_identifiers

    def clean(self):
        data = super().clean()

        if data.get("user_identifiers"):
            username_match = Q(slug__in=data["user_identifiers"])
            email_match = Q(email_hash__in=map(hash_email, data["user_identifiers"]))

            data["users"] = list(User.objects.filter(username_match | email_match))

            if len(data["users"]) != len(data["user_identifiers"]):
                raise forms.ValidationError(
                    _("One or more specified users could not be found.")
                )

        return data


class FilterDataDownloadsForm(forms.Form):
    status = forms.ChoiceField(
        label=_("Status"), required=False, choices=DataDownload.STATUS_CHOICES
    )
    user = forms.CharField(label=_("User"), required=False)
    requested_by = forms.CharField(label=_("Requested by"), required=False)

    def filter_queryset(self, criteria, queryset):
        if criteria.get("status") is not None:
            queryset = queryset.filter(status=criteria["status"])

        if criteria.get("user"):
            queryset = queryset.filter(user__slug__istartswith=criteria["user"])

        if criteria.get("requested_by"):
            queryset = queryset.filter(
                requester__slug__istartswith=criteria["requested_by"]
            )

        return queryset



from django.urls import reverse

from ...acl.test import patch_user_acl
from ...users.test import AuthenticatedUserTestCase
from ..searchproviders import searchproviders


class SearchApiTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.test_link = reverse("misago:api:search")

    @patch_user_acl({"can_search": False})
    def test_no_permission(self):
        """api validates permission to search"""
        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You don't have permission to search site."}
        )

    def test_no_phrase(self):
        """api handles no search query"""
        response = self.client.get(self.test_link)
        self.assertEqual(response.status_code, 200)

        providers = searchproviders.get_providers(True)
        for i, provider in enumerate(response.json()):
            provider_api = reverse(
                "misago:api:search", kwargs={"search_provider": providers[i].url}
            )
            self.assertEqual(provider_api, provider["api"])

            self.assertEqual(str(providers[i].name), provider["name"])
            self.assertEqual(provider["results"]["results"], [])
            self.assertEqual(int(provider["time"]), 0)

    def test_empty_search(self):
        """api handles empty search query"""
        response = self.client.get("%s?q=" % self.test_link)
        self.assertEqual(response.status_code, 200)

        providers = searchproviders.get_providers(True)
        for i, provider in enumerate(response.json()):
            provider_api = reverse(
                "misago:api:search", kwargs={"search_provider": providers[i].url}
            )
            self.assertEqual(provider_api, provider["api"])

            self.assertEqual(str(providers[i].name), provider["name"])
            self.assertEqual(provider["results"]["results"], [])
            self.assertEqual(int(provider["time"]), 0)



from django.conf import settings
from django.db import models
from django.utils import timezone


class PollVote(models.Model):
    category = models.ForeignKey("misago_categories.Category", on_delete=models.CASCADE)
    thread = models.ForeignKey("misago_threads.Thread", on_delete=models.CASCADE)
    poll = models.ForeignKey("misago_threads.Poll", on_delete=models.CASCADE)
    voter = models.ForeignKey(
        settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL
    )
    voter_name = models.CharField(max_length=255)
    voter_slug = models.CharField(max_length=255)
    voted_on = models.DateTimeField(default=timezone.now)
    choice_hash = models.CharField(max_length=12, db_index=True)

    class Meta:
        index_together = [["poll", "voter_name"]]



from django.apps import AppConfig


class MisagoReadTrackerConfig(AppConfig):
    name = "misago.readtracker"
    label = "misago_readtracker"
    verbose_name = "Misago Read Tracker"

    def ready(self):
        from . import signals as _



from django.test import RequestFactory
from django.urls import reverse

from .. import test
from ...cache.versions import get_cache_versions
from ...categories.models import Category
from ...conf.dynamicsettings import DynamicSettings
from ...users.test import AuthenticatedUserTestCase, create_test_user
from ..api.postendpoints.patch_post import patch_is_liked
from ..models import Post
from ..participants import (
    add_participant,
    change_owner,
    make_participants_aware,
    remove_participant,
    set_owner,
)


class AnonymizeEventsTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()

        category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category)

    def get_request(self, user=None):
        request = self.factory.get("/customer/details")
        request.user = user or self.user
        request.user_ip = "127.0.0.1"
        request.cache_versions = get_cache_versions()
        request.settings = DynamicSettings(request.cache_versions)
        request.include_frontend_context = False
        request.frontend_context = {}

        return request

    def test_anonymize_changed_owner_event(self):
        """changed owner event is anonymized by user.anonymize_data"""
        user = create_test_user("OtherUser", "otheruser@example.com")
        request = self.get_request()

        set_owner(self.thread, self.user)
        make_participants_aware(self.user, self.thread)
        change_owner(request, self.thread, user)

        user.anonymize_data(anonymous_username="Deleted")

        event = Post.objects.get(event_type="changed_owner")
        self.assertEqual(
            event.event_context,
            {
                "user": {
                    "id": None,
                    "username": user.username,
                    "url": reverse("misago:index"),
                }
            },
        )

    def test_anonymize_added_participant_event(self):
        """added participant event is anonymized by user.anonymize_data"""
        user = create_test_user("OtherUser", "otheruser@example.com")
        request = self.get_request()

        set_owner(self.thread, self.user)
        make_participants_aware(self.user, self.thread)
        add_participant(request, self.thread, user)

        user.anonymize_data(anonymous_username="Deleted")

        event = Post.objects.get(event_type="added_participant")
        self.assertEqual(
            event.event_context,
            {
                "user": {
                    "id": None,
                    "username": user.username,
                    "url": reverse("misago:index"),
                }
            },
        )

    def test_anonymize_owner_left_event(self):
        """owner left event is anonymized by user.anonymize_data"""
        user = create_test_user("OtherUser", "otheruser@example.com")
        request = self.get_request(user)

        set_owner(self.thread, user)
        make_participants_aware(user, self.thread)
        add_participant(request, self.thread, self.user)

        make_participants_aware(user, self.thread)
        remove_participant(request, self.thread, user)

        user.anonymize_data(anonymous_username="Deleted")

        event = Post.objects.get(event_type="owner_left")
        self.assertEqual(
            event.event_context,
            {
                "user": {
                    "id": None,
                    "username": user.username,
                    "url": reverse("misago:index"),
                }
            },
        )

    def test_anonymize_removed_owner_event(self):
        """removed owner event is anonymized by user.anonymize_data"""
        user = create_test_user("OtherUser", "otheruser@example.com")
        request = self.get_request()

        set_owner(self.thread, user)
        make_participants_aware(user, self.thread)
        add_participant(request, self.thread, self.user)

        make_participants_aware(user, self.thread)
        remove_participant(request, self.thread, user)

        user.anonymize_data(anonymous_username="Deleted")

        event = Post.objects.get(event_type="removed_owner")
        self.assertEqual(
            event.event_context,
            {
                "user": {
                    "id": None,
                    "username": user.username,
                    "url": reverse("misago:index"),
                }
            },
        )

    def test_anonymize_participant_left_event(self):
        """participant left event is anonymized by user.anonymize_data"""
        user = create_test_user("OtherUser", "otheruser@example.com")
        request = self.get_request(user)

        set_owner(self.thread, self.user)
        make_participants_aware(user, self.thread)
        add_participant(request, self.thread, user)

        make_participants_aware(user, self.thread)
        remove_participant(request, self.thread, user)

        user.anonymize_data(anonymous_username="Deleted")

        event = Post.objects.get(event_type="participant_left")
        self.assertEqual(
            event.event_context,
            {
                "user": {
                    "id": None,
                    "username": user.username,
                    "url": reverse("misago:index"),
                }
            },
        )

    def test_anonymize_removed_participant_event(self):
        """removed participant event is anonymized by user.anonymize_data"""
        user = create_test_user("OtherUser", "otheruser@example.com")
        request = self.get_request()

        set_owner(self.thread, self.user)
        make_participants_aware(self.user, self.thread)
        add_participant(request, self.thread, user)

        make_participants_aware(self.user, self.thread)
        remove_participant(request, self.thread, user)

        user.anonymize_data(anonymous_username="Deleted")

        event = Post.objects.get(event_type="removed_participant")
        self.assertEqual(
            event.event_context,
            {
                "user": {
                    "id": None,
                    "username": user.username,
                    "url": reverse("misago:index"),
                }
            },
        )


class AnonymizeLikesTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()

    def get_request(self, user=None):
        request = self.factory.get("/customer/details")
        request.user = user or self.user
        request.user_ip = "127.0.0.1"

        return request

    def test_anonymize_user_likes(self):
        """post's last like is anonymized by user.anonymize_data"""
        category = Category.objects.get(slug="first-category")
        thread = test.post_thread(category)
        post = test.reply_thread(thread)
        post.acl = {"can_like": True}

        user = create_test_user("OtherUser", "otheruser@example.com")

        patch_is_liked(self.get_request(self.user), post, 1)
        patch_is_liked(self.get_request(user), post, 1)

        user.anonymize_data(anonymous_username="Deleted")

        last_likes = Post.objects.get(pk=post.pk).last_likes
        self.assertEqual(
            last_likes,
            [
                {"id": None, "username": user.username},
                {"id": self.user.id, "username": self.user.username},
            ],
        )


class AnonymizePostsTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()

    def get_request(self, user=None):
        request = self.factory.get("/customer/details")
        request.user = user or self.user
        request.user_ip = "127.0.0.1"

        return request

    def test_anonymize_user_posts(self):
        """post is anonymized by user.anonymize_data"""
        category = Category.objects.get(slug="first-category")
        thread = test.post_thread(category)

        user = create_test_user("OtherUser", "otheruser@example.com")
        post = test.reply_thread(thread, poster=user)
        user.anonymize_data(anonymous_username="Deleted")

        anonymized_post = Post.objects.get(pk=post.pk)
        self.assertTrue(anonymized_post.is_valid)



from django import forms
from django.utils.translation import gettext_lazy as _

from ...acl import algebra
from ...acl.models import Role
from ...admin.forms import YesNoSwitch
from ..models import Attachment

# Admin Permissions Forms
class PermissionsForm(forms.Form):
    legend = _("Attachments")

    max_attachment_size = forms.IntegerField(
        label=_("Max attached file size (in kb)"),
        help_text=_("Enter 0 to don't allow uploading end deleting attachments."),
        initial=500,
        min_value=0,
    )

    can_download_other_users_attachments = YesNoSwitch(
        label=_("Can download other users attachments")
    )
    can_delete_other_users_attachments = YesNoSwitch(
        label=_("Can delete other users attachments")
    )


class AnonymousPermissionsForm(forms.Form):
    legend = _("Attachments")

    can_download_other_users_attachments = YesNoSwitch(
        label=_("Can download attachments")
    )


def change_permissions_form(role):
    if isinstance(role, Role):
        if role.special_role == "anonymous":
            return AnonymousPermissionsForm
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {
        "max_attachment_size": 0,
        "can_download_other_users_attachments": False,
        "can_delete_other_users_attachments": False,
    }
    new_acl.update(acl)

    return algebra.sum_acls(
        new_acl,
        roles=roles,
        key=key_name,
        max_attachment_size=algebra.greater,
        can_download_other_users_attachments=algebra.greater,
        can_delete_other_users_attachments=algebra.greater,
    )


def add_acl_to_attachment(user_acl, attachment):
    if user_acl["is_authenticated"] and user_acl["user_id"] == attachment.uploader_id:
        attachment.acl.update({"can_delete": True})
    else:
        user_can_delete = user_acl["can_delete_other_users_attachments"]
        attachment.acl.update(
            {"can_delete": user_acl["is_authenticated"] and user_can_delete}
        )


def register_with(registry):
    registry.acl_annotator(Attachment, add_acl_to_attachment)



from .cutoffdate import get_cutoff_date


def make_read_aware(request, posts):
    if not posts:
        return

    if not hasattr(posts, "__iter__"):
        posts = [posts]

    make_read(posts)

    if request.user.is_anonymous:
        return

    cutoff_date = get_cutoff_date(request.settings, request.user)
    unresolved_posts = {}

    for post in posts:
        if post.posted_on > cutoff_date:
            post.is_read = False
            post.is_new = True
            unresolved_posts[post.pk] = post

    if unresolved_posts:
        queryset = request.user.postread_set.filter(post__in=unresolved_posts)
        for post_id in queryset.values_list("post_id", flat=True):
            unresolved_posts[post_id].is_read = True
            unresolved_posts[post_id].is_new = False


def make_read(posts):
    for post in posts:
        post.is_read = True
        post.is_new = False


def save_read(user, post):
    user.postread_set.create(category=post.category, thread=post.thread, post=post)



import io

from PIL import Image
from django.core.files.images import ImageFile

from ...core.utils import get_file_hash

IMAGE_TYPES = ("image/gif", "image/png", "image/jpeg", "image/bmp", "image/webp")
THUMBNAIL_SIZE = (32, 32)


def create_media(theme, media):
    if media_exists(theme, media):
        delete_media(theme, media)

    if media_is_image(media):
        save_image(theme, media)
    else:
        save_media(theme, media)


def media_exists(theme, media):
    return theme.media.filter(name=media.name).exists()


def delete_media(theme, media):
    theme.media.get(name=media.name).delete()


def media_is_image(image):
    return image.content_type in IMAGE_TYPES


def save_image(theme, image):
    try:
        img = Image.open(image.file)
    except Exception:  # pylint: disable=broad-except
        return
    else:
        width, height = img.size

    theme.media.create(
        name=image.name,
        file=image,
        hash=get_file_hash(image),
        type=image.content_type,
        width=width,
        height=height,
        size=image.size,
        thumbnail=get_image_thumbnail(img, image.name),
    )

    img.close()


def get_image_thumbnail(image, src_name):
    img = image.copy()
    img.thumbnail(THUMBNAIL_SIZE)
    file = io.BytesIO()
    img.save(file, format="png")
    img.close()

    filename = ".".join(src_name.split(".")[:1])
    return ImageFile(file, name="thumb_%s.png" % filename)


def save_media(theme, media):
    theme.media.create(
        name=media.name,
        file=media,
        hash=get_file_hash(media),
        type=media.content_type,
        size=media.size,
    )



from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from rest_framework import viewsets
from rest_framework.response import Response

from ...core.shortcuts import get_int_or_404, paginate, pagination_dict
from ..models import UsernameChange
from ..serializers import UsernameChangeSerializer
from .rest_permissions import BasePermission

User = get_user_model()


class UsernameChangesViewSetPermission(BasePermission):
    def has_permission(self, request, view):
        try:
            user_pk = int(request.query_params.get("user"))
        except (ValueError, TypeError):
            user_pk = -1

        if user_pk == request.user.pk:
            return True
        if not request.user_acl.get("can_see_users_name_history"):
            raise PermissionDenied(
                _("You don't have permission to see other users name history.")
            )
        return True


class UsernameChangesViewSet(viewsets.GenericViewSet):
    permission_classes = (UsernameChangesViewSetPermission,)
    serializer_class = UsernameChangeSerializer

    def get_queryset(self):
        queryset = UsernameChange.objects

        if self.request.query_params.get("user"):
            user_pk = get_int_or_404(self.request.query_params.get("user"))
            queryset = get_object_or_404(User.objects, pk=user_pk).namechanges

        if self.request.query_params.get("search"):
            search_phrase = self.request.query_params.get("search").strip()
            if search_phrase:
                queryset = queryset.filter(
                    Q(changed_by_username__istartswith=search_phrase)
                    | Q(new_username__istartswith=search_phrase)
                    | Q(old_username__istartswith=search_phrase)
                )

        return queryset.select_related("user", "changed_by").order_by("-id")

    def list(self, request):
        page = get_int_or_404(request.query_params.get("page", 0))
        if page == 1:
            page = 0  # api allows explicit first page

        queryset = self.get_queryset()

        list_page = paginate(queryset, page, 12, 4)

        data = pagination_dict(list_page)
        data.update(
            {"results": UsernameChangeSerializer(list_page.object_list, many=True).data}
        )

        return Response(data)



from django.db import models
from django.utils.translation import gettext_lazy as _


class AttachmentType(models.Model):
    ENABLED = 0
    LOCKED = 1
    DISABLED = 2

    name = models.CharField(max_length=255)
    extensions = models.CharField(max_length=255)
    mimetypes = models.CharField(null=True, blank=True, max_length=255)
    size_limit = models.PositiveIntegerField(default=1024)
    status = models.PositiveIntegerField(
        default=ENABLED,
        choices=[
            (ENABLED, _("Allow uploads and downloads")),
            (LOCKED, _("Allow downloads only")),
            (DISABLED, _("Disallow both uploading and downloading")),
        ],
    )

    limit_uploads_to = models.ManyToManyField(
        "misago_acl.Role", related_name="+", blank=True
    )
    limit_downloads_to = models.ManyToManyField(
        "misago_acl.Role", related_name="+", blank=True
    )

    def __str__(self):
        return self.name

    @property
    def is_enabled(self):
        return self.status == AttachmentType.ENABLED

    @property
    def is_locked(self):
        return self.status == AttachmentType.LOCKED

    @property
    def extensions_list(self):
        if self.extensions:
            return self.extensions.split(",")
        return []

    @property
    def mimetypes_list(self):
        if self.mimetypes:
            return self.mimetypes.split(",")
        return []



from django import forms

from ..forms import YesNoSwitch


class YesNoForm(forms.Form):
    test_field = YesNoSwitch(label="Hello!")


def test_input_returns_int_true_for_valid_true_input():
    for value in ("1", "True", "true", 1, True):
        form = YesNoForm({"test_field": value})
        form.full_clean()
        assert form.cleaned_data["test_field"] == 1


def test_input_returns_int_false_for_false_input():
    for value in ("0", "False", "false", False, 0, None, ""):
        form = YesNoForm({"test_field": value})
        form.full_clean()
        assert form.cleaned_data["test_field"] == 0


def test_input_returns_int_false_for_invalid_input():
    form = YesNoForm({"test_field": "invalid"})
    form.full_clean()
    assert form.cleaned_data["test_field"] == 0



from django.test import TestCase

from ..utils import hash_email


class HashEmailTests(TestCase):
    def test_is_case_insensitive(self):
        """util is case insensitive"""
        self.assertEqual(hash_email("abc@test.com"), hash_email("aBc@tEst.cOm"))

    def test_handles_unicode(self):
        """util works with unicode strings"""
        self.assertEqual(hash_email("łóć@test.com"), hash_email("ŁÓĆ@tEst.cOm"))



import os
from glob import glob
from pathlib import Path

MISAGO_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
MOMENT_LOCALES_DIR = os.path.join(MISAGO_DIR, "static", "misago", "admin", "momentjs")


def get_admin_locales():
    locales = []
    for locale_path in glob(os.path.join(MOMENT_LOCALES_DIR, "*.js")):
        locales.append(Path(locale_path).stem)
    return locales


MISAGO_ADMIN_MOMENT_LOCALES = get_admin_locales()



class Providers:
    def __init__(self):
        self._dict = {}
        self._list = []

    def dict(self):
        return self._dict

    def list(self):
        return self._list

    def is_registered(self, provider):
        return provider in self._dict

    def add(
        self, *, provider, name, auth_backend, settings=None, admin_form, admin_template
    ):
        data = {
            "provider": provider,
            "name": name,
            "auth_backend": auth_backend,
            "settings": settings or {},
            "admin_form": admin_form,
            "admin_template": admin_template,
        }

        if provider != auth_backend.name:
            raise ValueError(
                f"Provider's key '{provider}' is not the same as backend's name "
                f"attribute: '{auth_backend.name}'"
            )

        self._dict[provider] = data
        self._list.append(data)
        self._list = sorted(self._list, key=lambda k: k["name"])

    def get_name(self, provider):
        return self._dict.get(provider)["name"]

    def get_auth_backend(self, provider):
        return self._dict.get(provider)["auth_backend"]

    def get_settings(self, provider):
        return self._dict.get(provider)["settings"]

    def get_admin_form_class(self, provider):
        return self._dict.get(provider)["admin_form"]

    def get_admin_template_name(self, provider):
        return self._dict.get(provider)["admin_template"]


providers = Providers()



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_finalization_sets_translation_strings_in_quotes 1"
] = '<div class="quote-heading">Quoted message:</div>'

snapshots[
    "test_finalization_sets_translation_strings_in_spoilers_buttons 1"
] = '<button class="spoiler-reveal" type="button">Reveal spoiler</button>'



import re

import markdown
from markdown.extensions.attr_list import AttrListExtension
from markdown.extensions.fenced_code import FencedBlockPreprocessor
from markdown.extensions.codehilite import CodeHilite, CodeHiliteExtension
from markdown.serializers import _escape_attrib_html


class CodeBlockExtension(markdown.Extension):
    def extendMarkdown(self, md):
        md.registerExtension(self)

        md.preprocessors.register(
            CodeBlockPreprocessor(md, self.getConfigs()), "misago_code_bbcode", 24
        )


class CodeBlockPreprocessor(FencedBlockPreprocessor):
    FENCED_BLOCK_RE = re.compile(
        r"""
\[code(=("?)(?P<lang>.*?)("?))?](([ ]*\n)+)?(?P<code>.*?)((\s|\n)+)?\[/code\]
""",
        re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE,
    )

    def run(self, lines):
        """Match and store Fenced Code Blocks in the HtmlStash."""

        # Check for dependent extensions
        if not self.checked_for_deps:
            for ext in self.md.registeredExtensions:
                if isinstance(ext, CodeHiliteExtension):
                    self.codehilite_conf = ext.getConfigs()
                if isinstance(ext, AttrListExtension):
                    self.use_attr_list = True

            self.checked_for_deps = True

        text = "\n".join(lines)
        while 1:
            m = self.FENCED_BLOCK_RE.search(text)
            if m:
                lang, id, classes, config = None, "", [], {}
                if m.group("lang"):
                    lang = m.group("lang")

                # If config is not empty, then the codehighlite extension
                # is enabled, so we call it to highlight the code
                if (
                    self.codehilite_conf
                    and self.codehilite_conf["use_pygments"]
                    and config.get("use_pygments", True)
                ):
                    local_config = self.codehilite_conf.copy()
                    local_config.update(config)
                    # Combine classes with cssclass. Ensure cssclass is at end
                    # as pygments appends a suffix under certain circumstances.
                    # Ignore ID as Pygments does not offer an option to set it.
                    if classes:
                        local_config["css_class"] = "{} {}".format(
                            " ".join(classes), local_config["css_class"]
                        )
                    highliter = CodeHilite(
                        m.group("code"),
                        lang=lang,
                        style=local_config.pop("pygments_style", "default"),
                        **local_config,
                    )

                    code = highliter.hilite(shebang=False)
                else:
                    id_attr = lang_attr = class_attr = kv_pairs = ""
                    if lang:
                        lang_attr = f' class="language-{_escape_attrib_html(lang)}"'
                    if classes:
                        class_attr = (
                            f' class="{_escape_attrib_html(" ".join(classes))}"'
                        )
                    if id:
                        id_attr = f' id="{_escape_attrib_html(id)}"'
                    if (
                        self.use_attr_list
                        and config
                        and not config.get("use_pygments", False)
                    ):
                        # Only assign key/value pairs to code element if attr_list ext is enabled, key/value pairs
                        # were defined on the code block, and the `use_pygments` key was not set to True. The
                        # `use_pygments` key could be either set to False or not defined. It is omitted from output.
                        kv_pairs = "".join(
                            f' {k}="{_escape_attrib_html(v)}"'
                            for k, v in config.items()
                            if k != "use_pygments"
                        )
                    code = self._escape(m.group("code"))
                    code = f"<pre{id_attr}{class_attr}><code{lang_attr}{kv_pairs}>{code}</code></pre>"

                placeholder = self.md.htmlStash.store(code)
                text = f"{text[:m.start()]}\n{placeholder}\n{text[m.end():]}"
            else:
                break
        return text.split("\n")



from django.utils.functional import SimpleLazyObject

from .enabledproviders import get_enabled_providers


def socialauth_providers_middleware(get_response):
    """Sets request.socialauth attribute with dict of setup social auth providers."""

    def middleware(request):
        def lazily_get_enabled_providers():
            return get_enabled_providers(request.cache_versions)

        request.socialauth = SimpleLazyObject(lazily_get_enabled_providers)
        return get_response(request)

    return middleware



from django.db import models


def create_audit_trail(request, obj):
    return create_user_audit_trail(request.user, request.user_ip, obj)


def create_user_audit_trail(user, ip_address, obj):
    if not isinstance(obj, models.Model):
        raise ValueError("obj must be a valid Django model instance")

    if user.is_anonymous:
        return None

    return user.audittrail_set.create(
        user=user, ip_address=ip_address, content_object=obj
    )



from datetime import timedelta

from django.utils import timezone

from ...conf import settings
from ..models import DataDownload
from ..signals import archive_user_data

from .dataarchive import DataArchive


STATUS_REQUEST = (DataDownload.STATUS_PENDING, DataDownload.STATUS_PROCESSING)


def user_has_data_download_request(user):
    queryset = DataDownload.objects.filter(user=user, status__in=STATUS_REQUEST)
    return queryset.exists()


def request_user_data_download(user, requester=None):
    requester = requester or user

    return DataDownload.objects.create(
        user=user, requester=requester, requester_name=requester.username
    )


def prepare_user_data_download(download, expires_in, logger=None):
    working_dir = settings.MISAGO_USER_DATA_DOWNLOADS_WORKING_DIR
    user = download.user
    with DataArchive(user, working_dir) as archive:
        try:
            archive_user_data.send(user, archive=archive)
            download.status = DataDownload.STATUS_READY
            download.expires_on = timezone.now() + timedelta(hours=expires_in)
            download.file = archive.get_file()
            download.save()
            # todo: send an e-mail with download link
            return True
        except Exception as e:  # pylint: disable=broad-except
            if logger:
                logger.exception(e, exc_info=e)
            return False


def expire_user_data_download(download):
    download.status = DataDownload.STATUS_EXPIRED
    if download.file:
        download.file.delete(save=False)
    download.save()



from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _

from ....acl.admin.forms import get_permissions_forms
from ....acl.admin.views import RoleAdmin, RolesList
from ....acl.cache import clear_acl_cache
from ....acl.models import Role
from ....admin.views import generic
from ...models import Category, CategoryRole, RoleCategoryACL
from ..forms import (
    CategoryRoleForm,
    CategoryRolesACLFormFactory,
    RoleCategoryACLFormFactory,
)
from .categories import CategoriesList, CategoryAdmin


class CategoryRoleAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:permissions:categories:index"
    model = CategoryRole
    templates_dir = "misago/admin/categoryroles"
    message_404 = _("Requested role does not exist.")


class CategoryRolesList(CategoryRoleAdmin, generic.ListView):
    ordering = (("name", None),)


class RoleFormMixin:
    def real_dispatch(self, request, target):
        form = CategoryRoleForm(instance=target)

        perms_forms = get_permissions_forms(target)

        if request.method == "POST":
            perms_forms = get_permissions_forms(target, request.POST)
            valid_forms = 0
            for permissions_form in perms_forms:
                if permissions_form.is_valid():
                    valid_forms += 1

            form = CategoryRoleForm(request.POST, instance=target)
            if form.is_valid():
                if len(perms_forms) == valid_forms:
                    new_permissions = {}
                    for permissions_form in perms_forms:
                        cleaned_data = permissions_form.cleaned_data
                        new_permissions[permissions_form.prefix] = cleaned_data

                    form.instance.permissions = new_permissions
                    form.instance.save()

                    messages.success(
                        request, self.message_submit % {"name": target.name}
                    )

                    if "stay" in request.POST:
                        return redirect(request.path)
                    return redirect(self.root_link)

                form.add_error(None, _("Form contains errors."))

        template_name = self.get_template_name(request, target)
        return self.render(
            request,
            {"form": form, "target": target, "perms_forms": perms_forms},
            template_name,
        )


class NewCategoryRole(RoleFormMixin, CategoryRoleAdmin, generic.ModelFormView):
    message_submit = _('New role "%(name)s" has been saved.')


class EditCategoryRole(RoleFormMixin, CategoryRoleAdmin, generic.ModelFormView):
    message_submit = _('Role "%(name)s" has been changed.')


class DeleteCategoryRole(CategoryRoleAdmin, generic.ButtonView):
    def check_permissions(self, request, target):
        if target.special_role:
            message = _('Role "%(name)s" is special role and can\'t be deleted.')
            return message % {"name": target.name}

    def button_action(self, request, target):
        target.delete()
        message = _('Role "%(name)s" has been deleted.')
        messages.success(request, message % {"name": target.name})


class CategoryPermissions(CategoryAdmin, generic.ModelFormView):
    templates_dir = "misago/admin/categoryroles"
    template_name = "categoryroles.html"

    def real_dispatch(self, request, target):
        category_roles = CategoryRole.objects.order_by("name")

        assigned_roles = {}
        for acl in target.category_role_set.select_related("category_role"):
            assigned_roles[acl.role_id] = acl.category_role

        forms = []
        forms_are_valid = True
        for role in Role.objects.order_by("name"):
            FormType = CategoryRolesACLFormFactory(
                role, category_roles, assigned_roles.get(role.pk)
            )

            if request.method == "POST":
                forms.append(FormType(request.POST, prefix=role.pk))
                if not forms[-1].is_valid():
                    forms_are_valid = False
            else:
                forms.append(FormType(prefix=role.pk))

        if request.method == "POST" and forms_are_valid:
            target.category_role_set.all().delete()
            new_permissions = []
            for form in forms:
                if form.cleaned_data["category_role"]:
                    new_permissions.append(
                        RoleCategoryACL(
                            role=form.role,
                            category=target,
                            category_role=form.cleaned_data["category_role"],
                        )
                    )
            if new_permissions:
                RoleCategoryACL.objects.bulk_create(new_permissions)

            clear_acl_cache()

            message = _("Category %(name)s permissions have been changed.")
            messages.success(request, message % {"name": target.name})
            if "stay" in request.POST:
                return redirect(request.path)
            return redirect(self.root_link)

        template_name = self.get_template_name(request, target)
        return self.render(request, {"forms": forms, "target": target}, template_name)


CategoriesList.add_item_action(
    name=_("Change permissions"), link="misago:admin:categories:permissions"
)


class RoleCategoriesACL(RoleAdmin, generic.ModelFormView):
    templates_dir = "misago/admin/categoryroles"
    template_name = "rolecategories.html"

    def real_dispatch(self, request, target):
        categories = Category.objects.all_categories()
        roles = CategoryRole.objects.order_by("name")

        if not categories:
            messages.info(request, _("No categories exist."))
            return redirect(self.root_link)

        choices = {}
        for choice in target.categories_acls.select_related("category_role"):
            choices[choice.category_id] = choice.category_role

        forms = []
        forms_are_valid = True
        for category in categories:
            category.level_range = range(category.level - 1)
            FormType = RoleCategoryACLFormFactory(
                category, roles, choices.get(category.pk)
            )

            if request.method == "POST":
                forms.append(FormType(request.POST, prefix=category.pk))
                if not forms[-1].is_valid():
                    forms_are_valid = False
            else:
                forms.append(FormType(prefix=category.pk))

        if request.method == "POST" and forms_are_valid:
            target.categories_acls.all().delete()
            new_permissions = []
            for form in forms:
                if form.cleaned_data["role"]:
                    new_permissions.append(
                        RoleCategoryACL(
                            role=target,
                            category=form.category,
                            category_role=form.cleaned_data["role"],
                        )
                    )
            if new_permissions:
                RoleCategoryACL.objects.bulk_create(new_permissions)

            clear_acl_cache()

            message = _("Category permissions for role %(name)s have been changed.")
            messages.success(request, message % {"name": target.name})
            if "stay" in request.POST:
                return redirect(request.path)
            return redirect(self.root_link)

        template_name = self.get_template_name(request, target)
        return self.render(request, {"forms": forms, "target": target}, template_name)


RolesList.add_item_action(
    name=_("Categories permissions"), link="misago:admin:permissions:categories"
)



from debug_toolbar.panels import Panel
from django.utils.translation import gettext_lazy as _


class MisagoACLPanel(Panel):
    """panel that displays current user's ACL"""

    title = _("Misago User ACL")
    template = "misago/acl_debug.html"

    @property
    def nav_subtitle(self):
        misago_user = self.get_stats().get("misago_user")

        if misago_user and misago_user.is_authenticated:
            return misago_user.username
        return _("Anonymous user")

    def generate_stats(self, request, response):
        try:
            misago_user = request.user
        except AttributeError:
            misago_user = None

        try:
            misago_acl = request.user_acl
        except AttributeError:
            misago_acl = {}

        self.record_stats({"misago_user": misago_user, "misago_acl": misago_acl})



from django import forms

from ..forms import ColorField


class ColorForm(forms.Form):
    test_field = ColorField(label="Hello!")


def test_input_returns_str_with_color_hex():
    form = ColorForm({"test_field": "#e9e9e9"})
    form.full_clean()
    assert form.cleaned_data["test_field"] == "#e9e9e9"


def test_input_rejects_short_color():
    form = ColorForm({"test_field": "#eee"})
    form.full_clean()
    assert "test_field" not in form.cleaned_data


def test_input_rejects_invalid_color():
    form = ColorForm({"test_field": "#ggg"})
    form.full_clean()
    assert "test_field" not in form.cleaned_data


def test_input_returns_none_for_empty_value():
    form = ColorForm({"test_field": ""})
    form.full_clean()
    assert "test_field" not in form.cleaned_data



from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.db.models import F
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework.response import Response

from ...acl.objectacl import add_acl_to_obj
from ...categories.models import Category
from ...core.rest_permissions import IsAuthenticatedOrReadOnly
from ...core.shortcuts import get_int_or_404
from ...threads.moderation import hide_post, hide_thread
from ..bans import get_user_ban
from ..datadownloads import request_user_data_download, user_has_data_download_request
from ..deletesrecord import record_user_deleted_by_staff
from ..online.utils import get_user_status
from ..permissions import (
    allow_browse_users_list,
    allow_delete_user,
    allow_edit_profile_details,
    allow_follow_user,
    allow_moderate_avatar,
    allow_rename_user,
    allow_see_ban_details,
)
from ..profilefields import profilefields, serialize_profilefields_data
from ..serializers import (
    BanDetailsSerializer,
    DataDownloadSerializer,
    DeleteOwnAccountSerializer,
    ForumOptionsSerializer,
    UserSerializer,
)
from ..viewmodels import Followers, Follows, UserPosts, UserThreads
from .rest_permissions import BasePermission, UnbannedAnonOnly
from .userendpoints.avatar import avatar_endpoint, moderate_avatar_endpoint
from .userendpoints.changeemail import change_email_endpoint
from .userendpoints.changepassword import change_password_endpoint
from .userendpoints.create import create_endpoint
from .userendpoints.editdetails import edit_details_endpoint
from .userendpoints.list import list_endpoint
from .userendpoints.signature import signature_endpoint
from .userendpoints.username import moderate_username_endpoint, username_endpoint

User = get_user_model()


class UserViewSetPermission(BasePermission):
    def has_permission(self, request, view):
        if view.action == "create":
            policy = UnbannedAnonOnly()
        else:
            policy = IsAuthenticatedOrReadOnly()
        return policy.has_permission(request, view)


def allow_self_only(user, pk, message):
    if user.is_anonymous:
        raise PermissionDenied(_("You have to sign in to perform this action."))
    if user.pk != int(pk):
        raise PermissionDenied(message)


class UserViewSet(viewsets.GenericViewSet):
    permission_classes = (UserViewSetPermission,)
    parser_classes = (FormParser, JSONParser, MultiPartParser)
    queryset = User.objects

    def get_queryset(self):
        relations = ("rank", "online_tracker", "ban_cache")
        return self.queryset.select_related(*relations)

    def get_user(self, request, pk):
        user = get_object_or_404(self.get_queryset(), pk=get_int_or_404(pk))
        if not user.is_active and not request.user.is_staff:
            raise Http404()
        return user

    def list(self, request):
        allow_browse_users_list(request.user_acl)
        return list_endpoint(request)

    def create(self, request):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                _(
                    "This feature has been disabled. "
                    "Please use %(provider)s to sign in."
                )
                % {"provider": request.settings.oauth2_provider}
            )

        return create_endpoint(request)

    def retrieve(self, request, pk=None):
        profile = self.get_user(request, pk)

        add_acl_to_obj(request.user_acl, profile)
        profile.status = get_user_status(request, profile)

        serializer = UserProfileSerializer(profile, context={"request": request})
        profile_json = serializer.data

        if not profile.is_active:
            profile_json["is_active"] = False
        if profile.is_deleting_account:
            profile_json["is_deleting_account"] = True

        return Response(profile_json)

    @action(methods=["get", "post"], detail=True)
    def avatar(self, request, pk=None):
        get_int_or_404(pk)
        allow_self_only(request.user, pk, _("You can't change other users avatars."))

        return avatar_endpoint(request)

    @action(
        methods=["post"],
        detail=True,
        url_name="forum-options",
        url_path="forum-options",
    )
    def forum_options(self, request, pk=None):
        get_int_or_404(pk)
        allow_self_only(request.user, pk, _("You can't change other users options."))

        serializer = ForumOptionsSerializer(request.user, data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"detail": _("Your forum options have been changed.")})
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    @action(methods=["get", "post"], detail=True)
    def username(self, request, pk=None):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                _(
                    "This feature has been disabled. "
                    "Please use %(provider)s to change your name."
                )
                % {"provider": request.settings.oauth2_provider}
            )

        get_int_or_404(pk)
        allow_self_only(request.user, pk, _("You can't change other users names."))

        return username_endpoint(request)

    @action(methods=["get", "post"], detail=True)
    def signature(self, request, pk=None):
        get_int_or_404(pk)
        allow_self_only(request.user, pk, _("You can't change other users signatures."))

        return signature_endpoint(request)

    @action(
        methods=["post"],
        detail=True,
        url_path="change-password",
        url_name="change-password",
    )
    def change_password(self, request, pk=None):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                _(
                    "This feature has been disabled. "
                    "Please use %(provider)s to change your password."
                )
                % {"provider": request.settings.oauth2_provider}
            )

        get_int_or_404(pk)
        allow_self_only(request.user, pk, _("You can't change other users passwords."))

        return change_password_endpoint(request)

    @action(
        methods=["post"], detail=True, url_path="change-email", url_name="change-email"
    )
    def change_email(self, request, pk=None):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                _(
                    "This feature has been disabled. "
                    "Please use %(provider)s to change your e-mail."
                )
                % {"provider": request.settings.oauth2_provider}
            )

        get_int_or_404(pk)
        allow_self_only(
            request.user, pk, _("You can't change other users e-mail addresses.")
        )

        return change_email_endpoint(request)

    @action(methods=["get"], detail=True)
    def details(self, request, pk=None):
        profile = self.get_user(request, pk)
        data = serialize_profilefields_data(request, profilefields, profile)
        return Response(data)

    @action(
        methods=["get", "post"],
        detail=True,
        url_path="edit-details",
        url_name="edit-details",
    )
    def edit_details(self, request, pk=None):
        profile = self.get_user(request, pk)
        allow_edit_profile_details(request.user_acl, profile)
        return edit_details_endpoint(request, profile)

    @action(
        methods=["post"],
        detail=True,
        url_path="delete-own-account",
        url_name="delete-own-account",
    )
    def delete_own_account(self, request, pk=None):
        if request.settings.enable_oauth2_client:
            raise PermissionDenied(
                _(
                    "This feature has been disabled. "
                    "Please use %(provider)s to delete your account."
                )
                % {"provider": request.settings.oauth2_provider}
            )

        serializer = DeleteOwnAccountSerializer(
            data=request.data, context={"user": request.user}
        )
        serializer.is_valid(raise_exception=True)
        serializer.mark_account_for_deletion(request)
        return Response({})

    @action(methods=["post"], detail=True)
    def follow(self, request, pk=None):
        profile = self.get_user(request, pk)
        allow_follow_user(request.user_acl, profile)

        profile_followers = profile.followers

        with transaction.atomic():
            if request.user.is_following(profile):
                request.user.follows.remove(profile)
                followed = False

                profile_followers -= 1
                profile.followers = F("followers") - 1
                request.user.following = F("following") - 1
            else:
                request.user.follows.add(profile)
                followed = True

                profile_followers += 1
                profile.followers = F("followers") + 1
                request.user.following = F("following") + 1

            profile.save(update_fields=["followers"])
            request.user.save(update_fields=["following"])

            return Response({"is_followed": followed, "followers": profile_followers})

    @action(detail=True)
    def ban(self, request, pk=None):
        profile = self.get_user(request, pk)
        allow_see_ban_details(request.user_acl, profile)

        ban = get_user_ban(profile, request.cache_versions)
        if ban:
            return Response(BanDetailsSerializer(ban).data)
        return Response({})

    @action(
        methods=["get", "post"],
        detail=True,
        url_path="moderate-avatar",
        url_name="moderate-avatar",
    )
    def moderate_avatar(self, request, pk=None):
        profile = self.get_user(request, pk)
        allow_moderate_avatar(request.user_acl, profile)

        return moderate_avatar_endpoint(request, profile)

    @action(
        methods=["get", "post"],
        detail=True,
        url_path="moderate-username",
        url_name="moderate-username",
    )
    def moderate_username(self, request, pk=None):
        profile = self.get_user(request, pk)
        allow_rename_user(request.user_acl, profile)

        return moderate_username_endpoint(request, profile)

    @action(
        methods=["post"],
        detail=True,
        url_path="request-data-download",
        url_name="request-data-download",
    )
    def request_data_download(self, request, pk=None):
        get_int_or_404(pk)
        allow_self_only(
            request.user, pk, _("You can't request data downloads for other users.")
        )

        if not request.settings.allow_data_downloads:
            raise PermissionDenied(_("You can't download your data."))

        if user_has_data_download_request(request.user):
            raise PermissionDenied(
                _(
                    "You can't have more than one data download request at a single time."
                )
            )

        request_user_data_download(request.user)

        return Response({"detail": "ok"})

    @action(methods=["get", "post"], detail=True)
    def delete(self, request, pk=None):
        profile = self.get_user(request, pk)
        allow_delete_user(request.user_acl, profile)

        if request.method == "POST":
            with transaction.atomic():
                profile.lock()

                if request.data.get("with_content"):
                    profile.delete_content()
                else:
                    categories_to_sync = set()

                    threads = profile.thread_set.select_related(
                        "category", "first_post"
                    )
                    for thread in threads.filter(is_hidden=False).iterator():
                        categories_to_sync.add(thread.category_id)
                        hide_thread(request, thread)

                    posts = profile.post_set.select_related(
                        "category", "thread", "thread__category"
                    )
                    for post in posts.filter(is_hidden=False).iterator():
                        categories_to_sync.add(post.category_id)
                        hide_post(request.user, post)
                        post.thread.synchronize()
                        post.thread.save()

                    categories = Category.objects.filter(id__in=categories_to_sync)
                    for category in categories.iterator():
                        category.synchronize()
                        category.save()

                profile.delete(anonymous_username=request.settings.anonymous_username)
                record_user_deleted_by_staff()

        return Response({})

    @action(
        methods=["get"],
        detail=True,
        url_path="data-downloads",
        url_name="data-downloads",
    )
    def data_downloads(self, request, pk=None):
        get_int_or_404(pk)
        allow_self_only(
            request.user, pk, _("You can't see other users data downloads.")
        )

        queryset = request.user.datadownload_set.all()[:5]
        serializer = DataDownloadSerializer(queryset, many=True)
        return Response(serializer.data)

    @action(methods=["get"], detail=True)
    def followers(self, request, pk=None):
        profile = self.get_user(request, pk)

        page = get_int_or_404(request.query_params.get("page", 0))
        if page == 1:
            page = 0  # api allows explicit first page

        search = request.query_params.get("search")

        users = Followers(request, profile, page, search)

        return Response(users.get_frontend_context())

    @action(methods=["get"], detail=True)
    def follows(self, request, pk=None):
        profile = self.get_user(request, pk)

        page = get_int_or_404(request.query_params.get("page", 0))
        if page == 1:
            page = 0  # api allows explicit first page

        search = request.query_params.get("search")

        users = Follows(request, profile, page, search)

        return Response(users.get_frontend_context())

    @action(methods=["get"], detail=True)
    def threads(self, request, pk=None):
        profile = self.get_user(request, pk)
        start = get_int_or_404(request.query_params.get("start", 0))
        feed = UserThreads(request, profile, start)
        return Response(feed.get_frontend_context())

    @action(methods=["get"], detail=True)
    def posts(self, request, pk=None):
        profile = self.get_user(request, pk)
        start = get_int_or_404(request.query_params.get("start", 0))
        feed = UserPosts(request, profile, start)
        return Response(feed.get_frontend_context())


UserProfileSerializer = UserSerializer.subset_fields(
    "id",
    "username",
    "slug",
    "email",
    "joined_on",
    "rank",
    "title",
    "avatars",
    "is_avatar_locked",
    "signature",
    "is_signature_locked",
    "followers",
    "following",
    "threads",
    "posts",
    "acl",
    "is_followed",
    "is_blocked",
    "real_name",
    "status",
    "api",
    "url",
)



from .models import DeletedUser


def record_user_deleted_by_self():
    return DeletedUser.objects.create(deleted_by=DeletedUser.DELETED_BY_SELF)


def record_user_deleted_by_staff():
    return DeletedUser.objects.create(deleted_by=DeletedUser.DELETED_BY_STAFF)


def record_user_deleted_by_system():
    return DeletedUser.objects.create(deleted_by=DeletedUser.DELETED_BY_SYSTEM)



from django.test import TestCase
from rest_framework import serializers

from ...categories.models import Category
from ...threads import test
from ...threads.models import Thread
from ..serializers import MutableFields


class Serializer(serializers.ModelSerializer, MutableFields):
    url = serializers.SerializerMethodField()

    class Meta:
        model = Thread
        fields = [
            "id",
            "title",
            "replies",
            "has_unapproved_posts",
            "started_on",
            "last_post_on",
            "last_post_is_event",
            "last_post",
            "last_poster_name",
            "is_unapproved",
            "is_hidden",
            "is_closed",
            "weight",
        ]


class MutableFieldsSerializerTests(TestCase):
    def test_subset_fields(self):
        """classmethod subset_fields creates new serializer"""
        category = Category.objects.get(slug="first-category")
        thread = test.post_thread(category=category)

        fields = ["id", "title", "replies", "last_poster_name"]

        serializer = Serializer.subset_fields(*fields)
        self.assertEqual(
            serializer.__name__, "SerializerIdTitleRepliesLastPosterNameSubset"
        )
        self.assertEqual(serializer.Meta.fields, fields)

        serialized_thread = serializer(thread).data
        self.assertEqual(
            serialized_thread,
            {
                "id": thread.id,
                "title": thread.title,
                "replies": thread.replies,
                "last_poster_name": thread.last_poster_name,
            },
        )

        self.assertFalse(Serializer.Meta.fields == serializer.Meta.fields)

    def test_exclude_fields(self):
        """classmethod exclude_fields creates new serializer"""
        category = Category.objects.get(slug="first-category")
        thread = test.post_thread(category=category)

        kept_fields = ["id", "title", "weight"]
        removed_fields = list(set(Serializer.Meta.fields) - set(kept_fields))

        serializer = Serializer.exclude_fields(*removed_fields)
        self.assertEqual(serializer.__name__, "SerializerIdTitleWeightSubset")
        self.assertEqual(serializer.Meta.fields, kept_fields)

        serialized_thread = serializer(thread).data
        self.assertEqual(
            serialized_thread,
            {"id": thread.id, "title": thread.title, "weight": thread.weight},
        )

        self.assertFalse(Serializer.Meta.fields == serializer.Meta.fields)

    def test_extend_fields(self):
        """classmethod extend_fields creates new serializer"""
        category = Category.objects.get(slug="first-category")
        thread = test.post_thread(category=category)

        serializer = Serializer.extend_fields("category")

        serialized_thread = serializer(thread).data
        self.assertEqual(serialized_thread["category"], category.pk)



from ..finalize import finalize_markup


def test_finalization_sets_translation_strings_in_quotes(snapshot):
    test_text = '<div class="quote-heading"></div>'
    finalized_text = finalize_markup(test_text)
    snapshot.assert_match(finalized_text)


def test_finalization_sets_translation_strings_in_spoilers_buttons(snapshot):
    test_text = '<button class="spoiler-reveal" type="button"></button>'
    finalized_text = finalize_markup(test_text)
    snapshot.assert_match(finalized_text)



from django.urls import reverse
from ...models import Setting
from ....test import ERROR, assert_has_message, assert_contains


def test_oauth2_can_be_enabled(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "1",
            "oauth2_provider": "Lorem",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": "",
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "avatar",
        },
    )

    assert response.status_code == 302

    settings = {row.setting: row.value for row in Setting.objects.all()}

    assert settings["enable_oauth2_client"] is True
    assert settings["oauth2_provider"] == "Lorem"
    assert settings["oauth2_client_id"] == "id"
    assert settings["oauth2_client_secret"] == "secret"
    assert settings["oauth2_scopes"] == "some scope"
    assert settings["oauth2_login_url"] == "https://example.com/login/"
    assert settings["oauth2_token_url"] == "https://example.com/token/"
    assert settings["oauth2_token_extra_headers"] == ""
    assert settings["oauth2_json_token_path"] == "access_token"
    assert settings["oauth2_user_url"] == "https://example.com/user/"
    assert settings["oauth2_user_method"] == "GET"
    assert settings["oauth2_user_token_location"] == "HEADER"
    assert settings["oauth2_user_token_name"] == "access_token"
    assert settings["oauth2_user_extra_headers"] == ""
    assert settings["oauth2_send_welcome_email"] is False
    assert settings["oauth2_json_id_path"] == "id"
    assert settings["oauth2_json_name_path"] == "name"
    assert settings["oauth2_json_email_path"] == "email"
    assert settings["oauth2_json_avatar_path"] == "avatar"


def test_oauth2_can_be_enabled_without_avatar(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "1",
            "oauth2_provider": "Lorem",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": "",
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert response.status_code == 302

    settings = {row.setting: row.value for row in Setting.objects.all()}

    assert settings["enable_oauth2_client"] is True
    assert settings["oauth2_provider"] == "Lorem"
    assert settings["oauth2_client_id"] == "id"
    assert settings["oauth2_client_secret"] == "secret"
    assert settings["oauth2_scopes"] == "some scope"
    assert settings["oauth2_login_url"] == "https://example.com/login/"
    assert settings["oauth2_token_url"] == "https://example.com/token/"
    assert settings["oauth2_token_extra_headers"] == ""
    assert settings["oauth2_json_token_path"] == "access_token"
    assert settings["oauth2_user_url"] == "https://example.com/user/"
    assert settings["oauth2_user_method"] == "GET"
    assert settings["oauth2_user_token_location"] == "HEADER"
    assert settings["oauth2_user_token_name"] == "access_token"
    assert settings["oauth2_user_extra_headers"] == ""
    assert settings["oauth2_send_welcome_email"] == False
    assert settings["oauth2_json_id_path"] == "id"
    assert settings["oauth2_json_name_path"] == "name"
    assert settings["oauth2_json_email_path"] == "email"
    assert settings["oauth2_json_avatar_path"] == ""


def test_oauth2_cant_be_enabled_with_some_value_missing(admin_client):
    data = {
        "enable_oauth2_client": "1",
        "oauth2_provider": "Lorem",
        "oauth2_client_id": "id",
        "oauth2_client_secret": "secret",
        "oauth2_scopes": "some scope",
        "oauth2_login_url": "https://example.com/login/",
        "oauth2_token_url": "https://example.com/token/",
        "oauth2_token_extra_headers": "",
        "oauth2_json_token_path": "access_token",
        "oauth2_user_url": "https://example.com/user/",
        "oauth2_user_method": "GET",
        "oauth2_user_token_location": "HEADER",
        "oauth2_user_token_name": "access_token",
        "oauth2_user_extra_headers": "",
        "oauth2_send_welcome_email": "",
        "oauth2_json_id_path": "id",
        "oauth2_json_name_path": "name",
        "oauth2_json_email_path": "email",
        "oauth2_json_avatar_path": "",
    }

    skip_settings = (
        "enable_oauth2_client",
        "oauth2_json_avatar_path",
        "oauth2_token_extra_headers",
        "oauth2_user_method",
        "oauth2_user_token_location",
        "oauth2_user_extra_headers",
        "oauth2_send_welcome_email",
    )

    for setting in data:
        if setting in skip_settings:
            continue

        new_data = data.copy()
        new_data[setting] = ""

        response = admin_client.post(
            reverse("misago:admin:settings:oauth2:index"),
            new_data,
        )

        assert response.status_code == 302
        assert_has_message(response, "You need to complete the configuration", ERROR)

        settings = {row.setting: row.value for row in Setting.objects.all()}

        assert settings["enable_oauth2_client"] is False

        if setting != "oauth2_client_id":
            assert settings["oauth2_client_id"] == "id"

        if setting != "oauth2_client_secret":
            assert settings["oauth2_client_secret"] == "secret"

        if setting != "oauth2_scopes":
            assert settings["oauth2_scopes"] == "some scope"

        if setting != "oauth2_login_url":
            assert settings["oauth2_login_url"] == "https://example.com/login/"

        if setting != "oauth2_token_url":
            assert settings["oauth2_token_url"] == "https://example.com/token/"

        if setting != "oauth2_json_token_path":
            assert settings["oauth2_json_token_path"] == "access_token"

        if setting != "oauth2_user_url":
            assert settings["oauth2_user_url"] == "https://example.com/user/"

        if setting != "oauth2_user_token_name":
            assert settings["oauth2_user_token_name"] == "access_token"

        if setting != "oauth2_json_id_path":
            assert settings["oauth2_json_id_path"] == "id"

        if setting != "oauth2_json_name_path":
            assert settings["oauth2_json_name_path"] == "name"

        if setting != "oauth2_json_email_path":
            assert settings["oauth2_json_email_path"] == "email"


def test_oauth2_scopes_are_normalized(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert response.status_code == 302

    setting = Setting.objects.get(setting="oauth2_scopes")
    assert setting.value == "some scope"


def test_oauth2_extra_token_headers_are_normalized(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": ("Lorem:   ipsum\n   Dolor: Met-elit"),
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": "",
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert response.status_code == 302

    setting = Setting.objects.get(setting="oauth2_token_extra_headers")
    assert setting.value == "Lorem: ipsum\nDolor: Met-elit"


def test_oauth2_extra_token_headers_are_validated(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": (
                "Lorem:   ipsum\n   Dolor-amet\n Dolor: Met-elit"
            ),
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": "",
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert_contains(response, "is not a valid header")


def test_oauth2_extra_user_headers_are_normalized(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": ("Lorem:   ipsum\n   Dolor: Met-amet"),
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert response.status_code == 302

    setting = Setting.objects.get(setting="oauth2_user_extra_headers")
    assert setting.value == "Lorem: ipsum\nDolor: Met-amet"


def test_oauth2_extra_user_headers_are_validated(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": ("Lorem:   ipsum\n   Dolor-met"),
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert_contains(response, "is not a valid header")


def test_oauth2_extra_headers_are_validated_to_have_colons(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": ("Lorem:   ipsum\n   Dolor-met"),
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert_contains(response, "is not a valid header. It&#x27;s missing a colon")


def test_oauth2_extra_headers_are_validated_to_have_names(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": ("Lorem:   ipsum\n   :Dolor-met"),
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert_contains(
        response,
        "is not a valid header. It&#x27;s missing a header name before the colon",
    )


def test_oauth2_extra_headers_are_validated_to_have_values(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": ("Lorem:   ipsum\n   Dolor-met:"),
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert_contains(
        response,
        "is not a valid header. It&#x27;s missing a header value after the colon",
    )


def test_oauth2_extra_headers_are_validated_to_be_unique(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:oauth2:index"),
        {
            "enable_oauth2_client": "0",
            "oauth2_client_id": "id",
            "oauth2_client_secret": "secret",
            "oauth2_scopes": "some some    scope",
            "oauth2_login_url": "https://example.com/login/",
            "oauth2_token_url": "https://example.com/token/",
            "oauth2_token_extra_headers": "",
            "oauth2_json_token_path": "access_token",
            "oauth2_user_url": "https://example.com/user/",
            "oauth2_user_method": "GET",
            "oauth2_user_token_location": "HEADER",
            "oauth2_user_token_name": "access_token",
            "oauth2_user_extra_headers": ("Accept:b\nLorem:   ipsum\n   Accept: a"),
            "oauth2_send_welcome_email": "",
            "oauth2_json_id_path": "id",
            "oauth2_json_name_path": "name",
            "oauth2_json_email_path": "email",
            "oauth2_json_avatar_path": "",
        },
    )

    assert_contains(response, "&quot;Accept&quot; header is entered more than once.")



from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from ..management.commands import loadavatargallery
from ..models import AvatarGallery


class LoadAvatarGalleryTests(TestCase):
    def test_regen_blank_avatar(self):
        """command regens blank avatar"""
        command = loadavatargallery.Command()

        out = StringIO()
        call_command(command, stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "New galleries have been loaded.")

        self.assertTrue(AvatarGallery.objects.exists())



import pytest
from django.urls import reverse

from ....test import assert_contains
from ...models import MenuItem


def test_nav_contains_menus_item(admin_client, list_url):
    response = admin_client.get(list_url)
    assert_contains(response, reverse("misago:admin:settings:menu-items:index"))


def test_empty_list_renders(admin_client, list_url):
    response = admin_client.get(list_url)
    assert response.status_code == 200


def test_list_renders_menu_item(admin_client, list_url, menu_item):
    response = admin_client.get(list_url)
    assert_contains(response, menu_item.title)


def test_menu_items_can_be_mass_deleted(admin_client, list_url, superuser):
    items = []
    for _ in range(10):
        item = MenuItem.objects.create(
            menu=MenuItem.MENU_FOOTER,
            title="Test Item {}".format(_),
            url="https://items{}.com".format(_),
        )
        items.append(item.pk)

    assert MenuItem.objects.count() == 10

    response = admin_client.post(
        list_url, data={"action": "delete", "selected_items": items}
    )
    assert response.status_code == 302
    assert MenuItem.objects.count() == 0


def test_creation_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:menu-items:new"))
    assert response.status_code == 200


def test_form_creates_new_menu_item(admin_client):
    response = admin_client.post(
        reverse("misago:admin:settings:menu-items:new"),
        {
            "menu": MenuItem.MENU_FOOTER,
            "title": "Test Item",
            "url": "https://admin.com/items/",
        },
    )

    item = MenuItem.objects.get()
    assert item.menu == MenuItem.MENU_FOOTER
    assert item.title == "Test Item"
    assert item.url == "https://admin.com/items/"


def test_edit_form_renders(admin_client, menu_item):
    response = admin_client.get(
        reverse("misago:admin:settings:menu-items:edit", kwargs={"pk": menu_item.pk})
    )
    assert_contains(response, menu_item.title)


def test_edit_form_updates_menu_items(admin_client, menu_item):
    response = admin_client.post(
        reverse("misago:admin:settings:menu-items:edit", kwargs={"pk": menu_item.pk}),
        data={
            "menu": menu_item.MENU_BOTH,
            "title": "Test Edited",
            "url": "https://example.com/edited/",
        },
    )
    assert response.status_code == 302

    menu_item.refresh_from_db()
    assert menu_item.menu == MenuItem.MENU_BOTH
    assert menu_item.title == "Test Edited"
    assert menu_item.url == "https://example.com/edited/"


def test_menu_item_can_be_deleted(admin_client, menu_item):
    response = admin_client.post(
        reverse("misago:admin:settings:menu-items:delete", kwargs={"pk": menu_item.pk})
    )
    assert response.status_code == 302

    with pytest.raises(MenuItem.DoesNotExist):
        menu_item.refresh_from_db()



from .forms import get_permissions_forms


def mock_role_form_data(model, data):
    """
    In order for form to don't fail submission, all permission fields need
    to receive values. This function populates data dict with default values
    for permissions, making form validation pass
    """
    for form in get_permissions_forms(model):
        for field in form:
            if field.value() is True:
                data[field.html_name] = 1
            elif field.value() is False:
                data[field.html_name] = 0
            else:
                data[field.html_name] = field.value()
    return data



from django.db import models
from django.utils.translation import gettext
from mptt.models import MPTTModel, TreeForeignKey

from .uploadto import (
    generate_theme_dirname,
    upload_build_css_to,
    upload_source_css_to,
    upload_media_to,
    upload_media_thumbnail_to,
)


class Theme(MPTTModel):
    parent = TreeForeignKey(
        "self", on_delete=models.PROTECT, null=True, blank=True, related_name="children"
    )
    name = models.CharField(max_length=255)
    dirname = models.CharField(max_length=8, default=generate_theme_dirname)
    is_default = models.BooleanField(default=False)
    is_active = models.BooleanField(default=False)

    version = models.CharField(max_length=255, null=True, blank=True)
    author = models.CharField(max_length=255, null=True, blank=True)
    url = models.URLField(max_length=255, null=True, blank=True)

    class MPTTMeta:
        order_insertion_by = ["-is_default", "name"]

    def delete(self, *args, **kwargs):
        for css in self.css.all():
            css.delete()
        for media in self.media.all():
            media.delete()

        super().delete(*args, **kwargs)

    def __str__(self):
        if self.is_default:
            return gettext("Default Misago Theme")
        return self.name

    @property
    def level_range(self):
        return range(self.level)


class Css(models.Model):
    theme = models.ForeignKey(Theme, on_delete=models.PROTECT, related_name="css")

    name = models.CharField(max_length=255)
    url = models.URLField(max_length=255, null=True, blank=True)
    source_file = models.FileField(
        upload_to=upload_source_css_to, max_length=255, null=True, blank=True
    )
    source_hash = models.CharField(max_length=8, null=True, blank=True)
    source_needs_building = models.BooleanField(default=False)
    build_file = models.FileField(
        upload_to=upload_build_css_to, max_length=255, null=True, blank=True
    )
    build_hash = models.CharField(max_length=8, null=True, blank=True)
    size = models.PositiveIntegerField(default=0)

    order = models.IntegerField(default=0)
    modified_on = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["order"]

    def delete(self, *args, **kwargs):
        if self.source_file:
            self.source_file.delete(save=False)
        if self.build_file:
            self.build_file.delete(save=False)

        super().delete(*args, **kwargs)

    def __str__(self):
        return self.name


class Media(models.Model):
    theme = models.ForeignKey(Theme, on_delete=models.PROTECT, related_name="media")

    name = models.CharField(max_length=255)
    file = models.ImageField(upload_to=upload_media_to, max_length=255)
    hash = models.CharField(max_length=8)
    type = models.CharField(max_length=255)
    width = models.PositiveIntegerField(default=0)
    height = models.PositiveIntegerField(default=0)
    size = models.PositiveIntegerField()
    thumbnail = models.ImageField(
        upload_to=upload_media_thumbnail_to, max_length=255, null=True, blank=True
    )
    modified_on = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["name"]

    def delete(self, *args, **kwargs):
        self.file.delete(save=False)
        if self.thumbnail:
            self.thumbnail.delete(save=False)
        super().delete(*args, **kwargs)

    def __str__(self):
        return self.name



from ...core.apirouter import MisagoApiRouter
from ..api import CategoryViewSet

router = MisagoApiRouter()
router.register(r"categories", CategoryViewSet, basename="category")
urlpatterns = router.urls



from django.contrib import messages
from django.db.models import Count
from django.utils.translation import gettext_lazy as _

from ....admin.views import generic
from ...models import AttachmentType
from ..forms import AttachmentTypeForm


class AttachmentTypeAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:settings:attachment-types:index"
    model = AttachmentType
    form_class = AttachmentTypeForm
    templates_dir = "misago/admin/attachmenttypes"
    message_404 = _("Requested attachment type could not be found.")

    def update_roles(self, target, roles):
        target.roles.clear()
        if roles:
            target.roles.add(*roles)

    def handle_form(self, form, request, target):
        super().handle_form(form, request, target)
        form.save()


class AttachmentTypesList(AttachmentTypeAdmin, generic.ListView):
    ordering = (("name", None),)

    def get_queryset(self):
        queryset = super().get_queryset()
        return queryset.annotate(num_files=Count("attachment"))


class NewAttachmentType(AttachmentTypeAdmin, generic.ModelFormView):
    message_submit = _('New type "%(name)s" has been saved.')


class EditAttachmentType(AttachmentTypeAdmin, generic.ModelFormView):
    message_submit = _('Attachment type "%(name)s" has been edited.')


class DeleteAttachmentType(AttachmentTypeAdmin, generic.ButtonView):
    def check_permissions(self, request, target):
        if target.attachment_set.exists():
            message = _(
                'Attachment type "%(name)s" has '
                "associated attachments and can't be deleted."
            )
            return message % {"name": target.name}

    def button_action(self, request, target):
        target.delete()
        message = _('Attachment type "%(name)s" has been deleted.')
        messages.success(request, message % {"name": target.name})



import random
from pathlib import Path

from django.core.files.base import ContentFile
from PIL import Image

from . import store
from ...conf import settings

DEFAULT_GALLERY = "__default__"


def get_available_galleries(include_default=False):
    """
    Returns list of dicts containing 'name' and list of images

    Only jpgs, gifs and pngs are supported avatar images.
    Galleries are
    """
    from ..models import AvatarGallery

    galleries = []
    galleries_dicts = {}

    for image in AvatarGallery.objects.all():
        if image.gallery == DEFAULT_GALLERY and not include_default:
            continue

        if image.gallery not in galleries_dicts:
            galleries_dicts[image.gallery] = {"name": image.gallery, "images": []}

            galleries.append(galleries_dicts[image.gallery])

        galleries_dicts[image.gallery]["images"].append(image)

    return galleries


def galleries_exist():
    from ..models import AvatarGallery

    return AvatarGallery.objects.exists()


def load_avatar_galleries():
    from ..models import AvatarGallery

    galleries = []
    for directory in Path(settings.MISAGO_AVATAR_GALLERY).iterdir():
        if not directory.is_dir():
            continue

        name = directory.name
        images = glob_gallery_images(directory)

        for image in images:
            with open(image, "rb") as image_file:
                galleries.append(
                    AvatarGallery.objects.create(
                        gallery=name, image=ContentFile(image_file.read(), "image")
                    )
                )
    return galleries


def glob_gallery_images(directory):
    images = []
    images.extend(directory.glob("*.gif"))
    images.extend(directory.glob("*.jpg"))
    images.extend(directory.glob("*.jpeg"))
    images.extend(directory.glob("*.png"))
    return images


def set_avatar(user, avatar):
    store.store_new_avatar(user, Image.open(avatar.image))


def set_random_avatar(user):
    galleries = get_available_galleries(include_default=True)
    if not galleries:
        raise RuntimeError("no avatar galleries are set")

    avatars_list = []
    for gallery in galleries:
        if gallery["name"] == DEFAULT_GALLERY:
            avatars_list = gallery["images"]
            break
        else:
            avatars_list += gallery["images"]

    random_avatar = random.choice(avatars_list)
    store.store_new_avatar(user, Image.open(random_avatar.image))



from django.dispatch import receiver

from ..users.signals import anonymize_user_data, username_changed
from .models import Agreement


@receiver([anonymize_user_data, username_changed])
def update_usernames(sender, **kwargs):
    Agreement.objects.filter(created_by=sender).update(created_by_name=sender.username)

    Agreement.objects.filter(last_modified_by=sender).update(
        last_modified_by_name=sender.username
    )



from django import forms

from ....cache.test import assert_invalidates_cache
from ... import SETTINGS_CACHE
from ..forms import ChangeSettingsForm


class Form(ChangeSettingsForm):
    settings = ["forum_name"]

    forum_name = forms.CharField(max_length=255)


def test_form_updates_setting_on_save(setting):
    form = Form({"forum_name": "New Value"}, request=None)
    assert form.is_valid()
    form.save({"forum_name": setting})

    setting.refresh_from_db()
    assert setting.value == "New Value"


def test_form_invalidates_settings_cache_on_save(setting):
    with assert_invalidates_cache(SETTINGS_CACHE):
        form = Form({"forum_name": "New Value"}, request=None)
        assert form.is_valid()
        form.save({"forum_name": setting})



from django.test import RequestFactory

from .. import test
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase, create_test_user
from ..api.postendpoints.patch_post import patch_is_liked
from ..models import Post


class DeleteUserLikesTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()

    def get_request(self, user=None):
        request = self.factory.get("/customer/details")
        request.user = user or self.user
        request.user_ip = "127.0.0.1"

        return request

    def test_anonymize_user_likes(self):
        """post's last like is anonymized by user.anonymize_data"""
        category = Category.objects.get(slug="first-category")
        thread = test.post_thread(category)
        post = test.reply_thread(thread)
        post.acl = {"can_like": True}

        user = create_test_user("OtherUser", "otheruser@example.com")

        patch_is_liked(self.get_request(self.user), post, 1)
        patch_is_liked(self.get_request(user), post, 1)

        user.delete_content()

        last_likes = Post.objects.get(pk=post.pk).last_likes
        self.assertEqual(
            last_likes, [{"id": self.user.id, "username": self.user.username}]
        )



from django.urls import reverse

from .. import test
from ...acl.test import patch_user_acl
from ..models import Thread, ThreadParticipant
from ..test import patch_private_threads_acl
from .test_privatethreads import PrivateThreadsTestCase


class PrivateThreadsListApiTests(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.api_link = reverse("misago:api:private-thread-list")

    def test_unauthenticated(self):
        """api requires user to sign in and be able to access it"""
        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have to sign in to use private threads."}
        )

    @patch_user_acl({"can_use_private_threads": False})
    def test_no_permission(self):
        """api requires user to have permission to be able to access it"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't use private threads."})

    @patch_user_acl({"can_use_private_threads": True})
    def test_empty_list(self):
        """api has no showstoppers on returning empty list"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(len(response_json["results"]), 0)

    @patch_user_acl({"can_use_private_threads": True})
    def test_thread_visibility(self):
        """only participated threads are returned by private threads api"""
        visible = test.post_thread(category=self.category, poster=self.user)
        reported = test.post_thread(category=self.category, poster=self.user)

        # hidden thread
        test.post_thread(category=self.category, poster=self.user)

        ThreadParticipant.objects.add_participants(visible, [self.user])

        reported.has_reported_posts = True
        reported.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(len(response_json["results"]), 1)
        self.assertEqual(response_json["results"][0]["id"], visible.id)

        # threads with reported posts will also show to moderators
        with patch_user_acl({"can_moderate_private_threads": True}):
            response = self.client.get(self.api_link)
            self.assertEqual(response.status_code, 200)

            response_json = response.json()
            self.assertEqual(len(response_json["results"]), 2)
            self.assertEqual(response_json["results"][0]["id"], reported.id)
            self.assertEqual(response_json["results"][1]["id"], visible.id)


class PrivateThreadRetrieveApiTests(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.thread = test.post_thread(self.category, poster=self.user)
        self.api_link = self.thread.get_api_url()

    def test_anonymous(self):
        """anonymous user can't see private thread"""
        self.logout_user()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have to sign in to use private threads."}
        )

    @patch_user_acl({"can_use_private_threads": False})
    def test_no_permission(self):
        """user needs to have permission to see private thread"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), {"detail": "You can't use private threads."})

    @patch_user_acl({"can_use_private_threads": True})
    def test_no_participant(self):
        """user cant see thread he isn't part of"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl(
        {"can_use_private_threads": True, "can_moderate_private_threads": True}
    )
    def test_mod_not_reported(self):
        """moderator can't see private thread that has no reports"""
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl(
        {"can_use_private_threads": True, "can_moderate_private_threads": False}
    )
    def test_reported_not_mod(self):
        """non-mod can't see private thread that has reported posts"""
        self.thread.has_reported_posts = True
        self.thread.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 404)

    @patch_user_acl({"can_use_private_threads": True})
    def test_can_see_owner(self):
        """user can see thread he is owner of"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["title"], self.thread.title)
        self.assertEqual(
            response_json["participants"],
            [
                {
                    "id": self.user.id,
                    "username": self.user.username,
                    "avatars": self.user.avatars,
                    "url": self.user.get_absolute_url(),
                    "is_owner": True,
                }
            ],
        )

    @patch_user_acl({"can_use_private_threads": True})
    def test_can_see_participant(self):
        """user can see thread he is participant of"""
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["title"], self.thread.title)
        self.assertEqual(
            response_json["participants"],
            [
                {
                    "id": self.user.id,
                    "username": self.user.username,
                    "avatars": self.user.avatars,
                    "url": self.user.get_absolute_url(),
                    "is_owner": False,
                }
            ],
        )

    @patch_user_acl(
        {"can_use_private_threads": True, "can_moderate_private_threads": True}
    )
    def test_mod_can_see_reported(self):
        """moderator can see private thread that has reports"""
        self.thread.has_reported_posts = True
        self.thread.save()

        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 200)

        response_json = response.json()
        self.assertEqual(response_json["title"], self.thread.title)
        self.assertEqual(response_json["participants"], [])


class PrivateThreadDeleteApiTests(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.thread = test.post_thread(self.category, poster=self.user)
        self.api_link = self.thread.get_api_url()

        ThreadParticipant.objects.add_participants(self.thread, [self.user])

    @patch_private_threads_acl({"can_hide_threads": 0})
    def test_hide_thread_no_permission(self):
        """api tests permission to delete threads"""

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json()["detail"], "You can't delete threads in this category."
        )

    @patch_private_threads_acl({"can_hide_threads": 1})
    def test_delete_thread_no_permission(self):
        """api tests permission to delete threads"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json()["detail"], "You can't delete threads in this category."
        )

    @patch_private_threads_acl({"can_hide_threads": 2})
    def test_delete_thread(self):
        """DELETE to API link with permission deletes thread"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)

        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)



from django.core.exceptions import ValidationError
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _
from django.utils.translation import ngettext

from .. import hooks
from ..categories import THREADS_ROOT_NAME
from ..categories.models import Category
from ..categories.permissions import can_browse_category, can_see_category
from ..conf import settings
from ..core.validators import validate_sluggable
from .threadtypes import trees_map


def validate_category(user_acl, category_id, allow_root=False):
    try:
        threads_tree_id = trees_map.get_tree_id_for_root(THREADS_ROOT_NAME)
        category = Category.objects.get(tree_id=threads_tree_id, id=category_id)
    except Category.DoesNotExist:
        category = None

    # Skip ACL validation for root category?
    if allow_root and category and not category.level:
        return category

    if not category or not can_see_category(user_acl, category):
        raise ValidationError(_("Requested category could not be found."))

    if not can_browse_category(user_acl, category):
        raise ValidationError(_("You don't have permission to access this category."))
    return category


def validate_thread_title(settings, title):
    validate_thread_title_length(settings, title)

    error_not_sluggable = _("Thread title should contain alpha-numeric characters.")
    error_slug_too_long = _("Thread title is too long.")
    validate_sluggable(error_not_sluggable, error_slug_too_long)(title)


def validate_thread_title_length(settings, value):
    value_len = len(value)

    if not value_len:
        raise ValidationError(_("You have to enter an thread title."))

    if value_len < settings.thread_title_length_min:
        # pylint: disable=line-too-long
        message = ngettext(
            "Thread title should be at least %(limit_value)s character long (it has %(show_value)s).",
            "Thread title should be at least %(limit_value)s characters long (it has %(show_value)s).",
            settings.thread_title_length_min,
        )
        raise ValidationError(
            message
            % {"limit_value": settings.thread_title_length_min, "show_value": value_len}
        )

    if value_len > settings.thread_title_length_max:
        # pylint: disable=line-too-long
        message = ngettext(
            "Thread title cannot be longer than %(limit_value)s character (it has %(show_value)s).",
            "Thread title cannot be longer than %(limit_value)s characters (it has %(show_value)s).",
            settings.thread_title_length_max,
        )
        raise ValidationError(
            message
            % {"limit_value": settings.thread_title_length_max, "show_value": value_len}
        )


def validate_post_length(settings, value):
    value_len = len(value)

    if not value_len:
        raise ValidationError(_("You have to enter a message."))

    if value_len < settings.post_length_min:
        # pylint: disable=line-too-long
        message = ngettext(
            "Posted message should be at least %(limit_value)s character long (it has %(show_value)s).",
            "Posted message should be at least %(limit_value)s characters long (it has %(show_value)s).",
            settings.post_length_min,
        )
        raise ValidationError(
            message % {"limit_value": settings.post_length_min, "show_value": value_len}
        )

    if settings.post_length_max and value_len > settings.post_length_max:
        # pylint: disable=line-too-long
        message = ngettext(
            "Posted message cannot be longer than %(limit_value)s character (it has %(show_value)s).",
            "Posted message cannot be longer than %(limit_value)s characters (it has %(show_value)s).",
            settings.post_length_max,
        )
        raise ValidationError(
            message % {"limit_value": settings.post_length_max, "show_value": value_len}
        )


# Post validation framework
validators_list = settings.MISAGO_POST_VALIDATORS
POST_VALIDATORS = list(map(import_string, validators_list))


def validate_post(context, data, validators=None):
    validators = validators or POST_VALIDATORS

    for validator in validators:
        data = validator(context, data) or data
    for validator in hooks.post_validators:
        data = validator(context, data) or data

    return data



from django.urls import path

from .views import oauth2_login, oauth2_complete

urlpatterns = [
    path("oauth2/login/", oauth2_login, name="oauth2-login"),
    path("oauth2/complete/", oauth2_complete, name="oauth2-complete"),
]
